from __future__ import annotations

import typing

import requests

import absorb

if typing.TYPE_CHECKING:
    from typing import Mapping, MutableMapping
    import polars as pl


class Chains(absorb.Table):
    source = 'chains'
    description = 'Registry of EVM chains and their chain IDs'
    url = 'https://github.com/ethereum-lists/chains'
    write_range = 'overwrite_all'
    index_type = 'number'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'name': pl.String,
            'chain_id': pl.Int64,
            'chain_id_hex': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        network_names = get_network_names()
        chain_ids = [int(chain_id) for chain_id in network_names.keys()]
        chain_id_hex = [
            hex(int(chain_id))[2:].rjust(64, '0') for chain_id in chain_ids
        ]
        data = {
            'name': network_names.values(),
            'chain_id': chain_ids,
            'chain_id_hex': chain_id_hex,
        }
        return pl.DataFrame(data, schema=self.get_schema())

    def get_available_range(self) -> absorb.Coverage:
        return (0, len(get_network_data()))


# specialcase the standard name for certain chains
special_cases: Mapping[str, str] = {
    'OP Mainnet': 'optimism',
    'Avalanche C-Chain': 'avalanche',
    'Arbitrum One': 'arbitrum',
    'BNB Smart Chain Mainnet': 'bsc',
    'Genesis Coin': 'genesis_coin',
    'X1 Network': 'x1_network',
    'ThaiChain 2.0 ThaiFi': 'thaifi',
    'WEMIX3.0 Mainnet': 'wemix',
    'WEMIX3.0 Testnet': 'wemix_testnet',
}


def get_network_names() -> Mapping[str, str]:
    # fetch raw network data
    data = get_network_data()
    network_names: MutableMapping[str, str] = {}
    for datum in data:
        # standardize name
        filtered = standardize_name(datum['name'])

        # skip deprecated networks
        if 'deprecated' in filtered:
            continue

        if filtered in network_names.values():
            continue
        network_names[str(datum['chainId'])] = filtered

    return network_names


def get_network_data() -> list[Mapping[str, typing.Any]]:
    url = 'https://chainid.network/chains.json'
    response = requests.get(url)
    return response.json()  # type: ignore


def standardize_name(name: str) -> str:
    """put name into standard format"""

    # special cases
    if name in special_cases:
        return special_cases[name]

    # replace special characters
    name = name.lower()
    name = name.replace(' ', '_')
    name = name.replace('-', '_')
    name = name.replace('___', '_')
    name = name.replace('__', '_')
    name = name.replace('(', '')
    name = name.replace(')', '')

    # remove keywords
    while True:
        remove = [
            '_mainnet',
            '_network',
            '_smart_chain',
            '_l1',
            '_sidechain',
            'sidechain',
            '_chain',
            'chain',
            '_coin',
        ]
        for piece in remove:
            if piece in name:
                name = name.replace(piece, '')
                break
        else:
            break

    # strip stray
    name = name.strip()
    name = name.strip('_')
    name = name.strip('-')

    return name



from __future__ import annotations

import typing

import absorb
from .. import cli_parsing
from . import command_path

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


cd_snippet_template = """function absorb {
    local tempfile="$(mktemp -t tmp.XXXXXX)"
    command absorb "$@" --cd-destination-tempfile "$tempfile"
    if [[ -s "$tempfile" ]]; then
        cd "$(realpath $(cat "$tempfile"))"
    fi
    rm -f "$tempfile" 2>/dev/null
}

function abs() {
    absorb "$@"
}
"""


def cd_command(args: Namespace) -> dict[str, Any]:
    # get path
    path = command_path._get_path(args)

    if args.cd_destination_tempfile is None:
        import sys

        print('using the cd subcommand requires special configuration')
        print()
        print(
            'add the following snippet to your shell config (e.g. ~/.profile):'
        )
        print()
        print(cd_snippet_template)
        sys.exit(0)

    # change pwd to path
    with open(args.cd_destination_tempfile, 'w') as f:
        f.write(path)

    return {}



from __future__ import annotations

import os
import typing
import absorb

from . import names

if typing.TYPE_CHECKING:
    import types
    import polars as pl


def get_source_module(source: str) -> types.ModuleType:
    import importlib

    return importlib.import_module(
        'absorb.catalog.' + names._camel_to_snake(source)
    )


def get_table_class(
    *,
    source: str | None = None,
    table_name: str | None = None,
    class_path: str | None = None,
) -> type[absorb.Table]:
    if class_path is not None:
        import importlib

        module_name, class_name = class_path.rsplit('.', maxsplit=1)
        module = importlib.import_module(module_name)
        return getattr(module, class_name)  # type: ignore
    elif source is not None and table_name is not None:
        return getattr(  # type: ignore
            get_source_module(source), names._snake_to_camel(table_name)
        )
    else:
        raise Exception(
            'either specify class_path or both source and table_name'
        )


def get_sources(*, snake: bool = True) -> list[str]:
    import absorb.catalog

    sources = [
        filename.rsplit('.py', maxsplit=1)[0]
        for filename in os.listdir(absorb.catalog.__path__[0])
        if not filename.startswith('__')
    ]

    if not snake:
        sources = [names._snake_to_camel(source) for source in sources]

    return sources


def get_source_table_classes(source: str) -> list[type[absorb.Table]]:
    module = get_source_module(source)
    if hasattr(module, 'get_tables'):
        return module.get_tables()  # type: ignore
    else:
        return [
            value
            for key, value in vars(module).items()
            if isinstance(value, type) and issubclass(value, absorb.Table)
        ]


def get_table_classes() -> list[type[absorb.Table]]:
    return [
        table_class
        for source in get_sources()
        for table_class in get_source_table_classes(source)
    ]



from __future__ import annotations

import typing

import absorb
from . import config

if typing.TYPE_CHECKING:
    import polars as pl


def check_bucket_setup(bucket: absorb.Bucket | None = None) -> str | None:
    # check if rclone package is installed
    if not config.is_package_installed('rclone_python'):
        return 'rclone_python is not installed. Install it before using buckets (for example, `uv add rclone_python`)'

    # check if rclone is installed
    import rclone_python.rclone  # type: ignore

    if not rclone_python.rclone.is_installed():
        return 'rclone is not installed. Install it before using buckets (for example, `brew install rclone`)'

    # check that remote is setup
    remotes = rclone_python.rclone.get_remotes()
    if len(remotes) == 0:
        return 'No rclone remotes are configured. Configure a remote using `rclone config` on the command line'

    if bucket is not None:
        for key in ['rclone_remote', 'bucket_name', 'path_prefix']:
            if key not in bucket:
                return f'Bucket configuration is missing required key: {key}'
        if not rclone_python.rclone.check_remote_existing(
            bucket['rclone_remote']
        ):
            return (
                'rclone remote '
                + str(bucket['rclone_remote'])
                + ' does not exist. Check your rclone configuration.'
            )

    return None


def get_default_bucket() -> absorb.Bucket:
    return absorb.ops.get_config()['default_bucket']


#
# # bucket scanning
#


def scan_bucket(
    table: absorb.TableReference,
    bucket: absorb.Bucket | None = None,
    scan_kwargs: dict[str, typing.Any] | None = None,
    verbose: bool = True,
) -> pl.LazyFrame:
    import polars as pl

    glob = get_table_bucket_glob(bucket=bucket, table=table)
    if scan_kwargs is None:
        scan_kwargs = {}
    if verbose:
        print('scanning remote bucket:', glob)
    return pl.scan_parquet(glob, **scan_kwargs)


def get_table_bucket_glob(
    table: absorb.TableReference,
    bucket: absorb.Bucket | None = None,
) -> str:
    # determine bucket
    if bucket is None:
        bucket = get_default_bucket()

    # get bucket protocol
    if bucket['provider'] == 'gcp':
        protocol = 'gs'
    elif bucket['provider'] == 'aws':
        protocol = 's3'
    else:
        raise Exception()
    bucket_name = bucket['bucket_name']
    if bucket_name is None:
        raise Exception('bucket must be specified')
    path_prefix = bucket.get('path_prefix')
    if path_prefix is None:
        raise Exception('path_prefix must be specified')

    # resolve table
    table = absorb.Table.instantiate(table)

    return (
        protocol
        + '://'
        + bucket_name
        + '/'
        + path_prefix
        + '/datasets/'
        + table.source
        + '/tables/'
        + table.name()
        + '/*.parquet'
    )


#
# # uploads/downloads
#


def upload_tables_to_bucket(
    tables: typing.Sequence[absorb.TableReference],
    bucket: absorb.Bucket | None = None,
) -> None:
    # determine bucket
    if bucket is None:
        bucket = get_default_bucket()

    # check bucket setup
    problem = absorb.ops.check_bucket_setup(bucket=bucket)
    if problem is not None:
        raise Exception(problem)

    # upload each bucket
    for table in tables:
        _upload_table_to_bucket(table, bucket)


def download_tables_to_bucket(
    tables: typing.Sequence[absorb.TableReference],
    bucket: absorb.Bucket | None = None,
) -> None:
    # determine bucket
    if bucket is None:
        bucket = get_default_bucket()

    # check bucket setup
    problem = absorb.ops.check_bucket_setup(bucket=bucket)
    if problem is not None:
        raise Exception(problem)

    # upload each bucket
    for table in tables:
        _download_table_from_bucket(table, bucket)


def _upload_table_to_bucket(
    table: absorb.TableReference,
    bucket: absorb.Bucket,
) -> None:
    import rclone_python.rclone

    # get paths
    table = absorb.Table.instantiate(table)
    table_dir = table.get_table_dir()
    bucket_path = get_rclone_bucket_path(table=table, **bucket)

    print('uploading', table_dir, 'to', bucket_path)

    # perform upload
    rclone_python.rclone.copy(table_dir, bucket_path)


def _download_table_from_bucket(
    table: absorb.TableReference,
    bucket: absorb.Bucket,
) -> None:
    import rclone_python.rclone

    # get paths
    table = absorb.Table.instantiate(table)
    table_dir = table.get_table_dir()
    bucket_path = get_rclone_bucket_path(table=table, **bucket)

    # perform upload
    rclone_python.rclone.copy(bucket_path, table_dir)


def get_rclone_bucket_path(
    *,
    rclone_remote: str | None,
    bucket_name: str | None,
    path_prefix: str | None,
    table: absorb.Table,
    provider: str | None = None,
) -> str:
    if rclone_remote is None:
        raise Exception('rclone_remote must be specified')
    if bucket_name is None:
        raise Exception('bucket_name must be specified')
    if path_prefix is None:
        raise Exception('path_prefix must be specified')
    return (
        rclone_remote.strip('/')
        + ':'
        + bucket_name.strip('/')
        + '/'
        + path_prefix.strip('/')
        + '/datasets/'
        + table.source
        + '/tables/'
        + table.name()
    )



from __future__ import annotations

import typing

import absorb
from . import common


if typing.TYPE_CHECKING:
    import polars as pl


class Yields(absorb.Table):
    source = 'defillama'
    description = 'Yields of each pool in USD over time'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'pools': (list, type(None)), 'top_n': int}
    default_parameters = {'pools': None, 'top_n': 5000}
    index_type = 'day'
    name_template = [
        'pool_yields_top_{top_n}',
        'pool_yields_{pools}',
        'pool_yields',
    ]

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'pool': pl.String,
            'chain': pl.String,
            'project': pl.String,
            'symbol': pl.String,
            'tvl_usd': pl.Float64,
            'apy_base': pl.Float64,
            'apy_base_7d': pl.Float64,
            'apy_reward': pl.Float64,
            'il_7d': pl.Float64,
            'revenue': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import time
        import polars as pl

        # get current yield pools
        current_yields = get_current_yields()

        # select which pools to collect
        pools = self.parameters['pools']
        if pools is None:
            pools = current_yields.sort('tvl_usd', descending=True)['pool']
            if self.parameters['top_n'] is not None:
                pools = pools[: self.parameters['top_n']]

        # collect yields for each pool
        next_time = time.time()
        dfs = []
        print('collecting', len(pools), 'pools')
        for p, pool in enumerate(pools, start=1):
            while time.time() < next_time:
                time.sleep(0.05)
            print('[' + str(p) + ' / ' + str(len(pools)) + ']', pool)
            try:
                df = get_historical_yields_of_pool(pool)
                dfs.append(df)
            except Exception:
                pass
            next_time = next_time + 4.0
        df = pl.concat(dfs)

        # get labels
        current_yields = current_yields[['pool', 'chain', 'project', 'symbol']]

        return df.join(current_yields, on='pool', how='left').select(
            self.get_schema().keys()
        )


def get_current_yields() -> pl.DataFrame:
    import polars as pl

    data = common._fetch('current_yields')

    columns = {
        'pool': 'pool',
        'chain': 'chain',
        'project': 'project',
        'symbol': 'symbol',
        'tvl_usd': 'tvlUsd',
        'apy_base': 'apyBase',
        'apy_reward': 'apyReward',
        'apy': 'apy',
        'apy_pct_1D': 'apyPct1D',
        'apy_pct_7D': 'apyPct7D',
        'apy_pct_30D': 'apyPct30D',
        'apy_mean_30d': 'apyMean30d',
        'apy_base_7d': 'apyBase7d',
        'apy_base_inception': 'apyBaseInception',
        'volume_usd_1d': 'volumeUsd1d',
        'volume_usd_7d': 'volumeUsd7d',
        'reward_tokens': 'rewardTokens',
        'underlying_tokens': 'underlyingTokens',
        'stablecoin': 'stablecoin',
        'il_risk': 'ilRisk',
        'il_7d': 'il7d',
        'exposure': 'exposure',
        'pool_meta': 'poolMeta',
        'mu': 'mu',
        'sigma': 'sigma',
        'count': 'count',
        'outlier': 'outlier',
    }

    return (
        pl.DataFrame(data['data'], infer_schema_length=99999999)
        .select(**columns)
        .sort('tvl_usd', descending=True)
    )


def get_historical_yields_of_pool(pool: str) -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_yields_per_pool', {'pool': pool})
    columns: dict[str, str | pl.Expr] = {
        'timestamp': pl.col.timestamp.str.to_datetime(),
        'pool': pl.lit(pool, dtype=pl.String),
        'tvl_usd': 'tvlUsd',
        'apy_base': 'apyBase',
        'apy_base_7d': 'apyBase7d',
        'apy_reward': 'apyReward',
        'il_7d': 'il7d',
    }
    raw_schema = {
        'timestamp': pl.String,
        'tvlUsd': pl.Float64,
        'apyBase': pl.Float64,
        'apyBase7d': pl.Float64,
        'apyReward': pl.Float64,
        'il7d': pl.Float64,
    }
    return (
        pl.DataFrame(data['data'], schema=raw_schema)
        .select(**columns)
        .with_columns(
            revenue=pl.col.tvl_usd * (pl.col.apy_base + pl.col.apy_reward),
        )
    )



from __future__ import annotations

import typing
import absorb

from . import table_coverage

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')

    import polars as pl


class TableCollect(table_coverage.TableCoverage):
    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        raise NotImplementedError()

    def is_collected(self) -> bool:
        import glob

        data_glob = self.get_glob()
        return len(glob.glob(data_glob)) > 0

    def collect(
        self,
        data_range: typing.Any | None = None,
        *,
        overwrite: bool = False,
        verbose: int = 1,
        dry: bool = False,
    ) -> None:
        self._check_ready_to_collect()

        # get collection plan
        chunks = self._get_chunks_to_collect(data_range, overwrite)

        # summarize collection plan
        if verbose >= 1:
            self._summarize_collection_plan(chunks, overwrite, verbose, dry)

        # return early if dry
        if dry:
            return None

        # create table directory
        self.setup_table_dir()

        # collect each chunk
        for chunk in chunks:
            self._execute_collect_chunk(chunk, overwrite, verbose)

    def _check_ready_to_collect(self) -> None:
        import os

        # check package dependencies
        missing_packages = self.get_missing_packages()
        if len(missing_packages) > 0:
            raise Exception(
                'required packages not installed: '
                + ', '.join(missing_packages)
            )

        # check credentials
        missing_credentials = self.get_missing_credentials()
        if len(missing_credentials) > 0:
            raise Exception(
                'required credentials not found: '
                + ', '.join(missing_credentials)
            )

    def _get_chunks_to_collect(
        self, data_range: absorb.Coverage | None = None, overwrite: bool = False
    ) -> absorb.ChunkList:
        if self.write_range == 'overwrite_all':
            available_range = self.get_available_range()
            collected_range = self.get_collected_range()
            if available_range is not None:
                # if available range is not None, use it to check for updates
                if available_range == collected_range:
                    return []
                else:
                    return [available_range]
            elif collected_range is not None and absorb.ops.index_is_temporal(
                self.index_type
            ):
                # if available range is None, check if ready for update
                if self.ready_for_update():
                    return [None]
                else:
                    return []
            else:
                # if a non-temporal index, collect entire dataset
                return [available_range]

        else:
            index_type = self.index_type
            if index_type is None:
                raise Exception(
                    'index type is required if not using overwrite_all'
                )

            # get coverage range for collection
            if data_range is None:
                if overwrite:
                    coverage = self.get_available_range()
                else:
                    coverage = self.get_missing_ranges()
            else:
                coverage = [data_range]
            if coverage is None:
                raise Exception()

            # partition coverage into chunks
            data_ranges = absorb.ops.coverage_to_list(
                coverage, index_type=index_type
            )
            return absorb.ops.partition_into_chunks(
                data_ranges, index_type=index_type
            )

    def _summarize_collection_plan(
        self,
        chunks: absorb.ChunkList,
        overwrite: bool,
        verbose: int,
        dry: bool,
    ) -> None:
        import datetime
        import rich

        rich.print(
            '[bold][green]collecting dataset:[/green] [white]'
            + self.full_name()
            + '[/white][/bold]'
        )
        absorb.ops.print_bullet('n_chunks', str(len(chunks)))
        if self.write_range == 'overwrite_all':
            absorb.ops.print_bullet('chunk', '\\' + '[entire dataset]')  # noqa
        elif len(chunks) == 1:
            absorb.ops.print_bullet(
                'single chunk',
                absorb.ops.format_chunk(chunks[0], self.index_type),
            )
        elif len(chunks) > 1:
            absorb.ops.print_bullet(
                'min_chunk',
                absorb.ops.format_chunk(chunks[0], self.index_type),
                indent=4,
            )
            absorb.ops.print_bullet(
                'max_chunk',
                absorb.ops.format_chunk(chunks[-1], self.index_type),
                indent=4,
            )
        absorb.ops.print_bullet('overwrite', str(overwrite))
        absorb.ops.print_bullet('output dir', self.get_table_dir())
        absorb.ops.print_bullet(
            'collection start time', str(datetime.datetime.now())
        )
        if len(chunks) == 0:
            print('[already collected]')

        if verbose > 1:
            print()
            absorb.ops.print_bullet(key='chunks', value='')
            for c, chunk in enumerate(chunks):
                absorb.ops.print_bullet(
                    key=None,
                    value=absorb.ops.format_chunk(chunk, self.index_type),
                    number=c + 1,
                    indent=4,
                )

        if dry:
            print('[dry run]')
        if len(chunks) > 0:
            print()

    def _execute_collect_chunk(
        self,
        chunk: absorb.Chunk,
        overwrite: bool,
        verbose: int,
    ) -> None:
        import os

        # print summary
        if verbose >= 1:
            if self.write_range == 'overwrite_all':
                as_str = 'all'
            else:
                as_str = absorb.ops.format_chunk(chunk, self.index_type)
            print('collecting', as_str)

        # collect chunk
        data = self.collect_chunk(chunk=chunk)

        # validate chunk
        self.validate_chunk(chunk=chunk, data=data)

        # write file
        if self.chunk_datatype == 'dataframe' and data is not None:
            import polars as pl

            if not isinstance(data, pl.DataFrame):
                raise Exception(
                    'collected data is not a DataFrame: ' + str(type(data))
                )
            path = self.get_file_path(chunk=chunk, df=data)
            absorb.ops.write_file(df=data, path=path)

        # print post-summary
        if verbose >= 1 and data is None:
            print('could not collect data for', str(chunk))

    def validate_chunk(
        self, chunk: absorb.Chunk, data: absorb.ChunkData | None
    ) -> None:
        import os
        import polars as pl

        if data is None:
            return

        if self.chunk_datatype == 'dataframe':
            if not isinstance(data, pl.DataFrame):
                raise Exception(
                    'collected data is not a DataFrame: ' + str(type(data))
                )
            assert dict(data.schema) == self.get_schema(), (
                'collected data does not match schema: '
                + str(dict(data.schema))
                + ' != '
                + str(self.get_schema())
            )
        elif self.chunk_datatype == 'files':
            if not isinstance(data, dict) or data.get('type') != 'files':
                raise Exception(
                    'collected data is not a path dict: ' + str(type(data))
                )
            for path in data['paths']:
                assert os.path.exists(path), (
                    'collected data does not exist: ' + path
                )
                file_schema = pl.scan_parquet(path).collect_schema()
                assert dict(file_schema) == self.get_schema(), (
                    'collected data does not match schema: '
                    + str(dict(file_schema))
                    + ' != '
                    + str(self.get_schema())
                )
        else:
            raise Exception('invalid data format: ' + str(type(data)))



from __future__ import annotations

import typing

import absorb
from . import snowflake

if typing.TYPE_CHECKING:
    import polars as pl


class Query(snowflake.Query):
    source = 'allium'
    url = 'https://docs.allium.so/historical-data/overview'


class StablecoinSupply(Query):
    description = 'Stablecoin supply data from Allium'
    url = 'https://docs.allium.so/historical-data/stablecoins'

    sql = 'SELECT * FROM CROSSCHAIN_ALLIUM.STABLECOIN.SUPPLY_BETA'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'DATE': pl.Datetime(time_unit='ns', time_zone=None),
            'CHAIN': pl.String,
            'TOKEN_ADDRESS': pl.String,
            'TOKEN_SYMBOL': pl.String,
            'TOKEN_NAME': pl.String,
            'BASE_ASSET': pl.String,
            'SUPPLY': pl.Float64,
            'SUPPLY_USD': pl.Float64,
            'SUPPLY_DELTA': pl.Float64,
            'SUPPLY_DELTA_USD': pl.Float64,
            'IS_BRIDGE': pl.Boolean,
            'CURRENCY': pl.String,
            'UNIQUE_ID': pl.String,
            '_CREATED_AT': pl.Datetime(time_unit='ns', time_zone=None),
            '_UPDATED_AT': pl.Datetime(time_unit='ns', time_zone=None),
            '_CHANGED_SINCE_FULL_REFRESH': pl.Boolean,
        }



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl

url_template = (
    'https://archive.blocknative.com/{year}{month:02}{day:02}/{hour:02}.csv.gz'
)


class Mempool(absorb.Table):
    source = 'blocknative'
    description = 'Snapshots of the Ethereum mempool'
    url = 'https://docs.blocknative.com/data-archive/mempool-archive'
    write_range = 'overwrite_all'
    index_type = 'hour'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'detecttime': pl.String,
            'hash': pl.String,
            'status': pl.String,
            'region': pl.String,
            'reorg': pl.String,
            'replace': pl.String,
            'curblocknumber': pl.Int64,
            'failurereason': pl.String,
            'blockspending': pl.Int64,
            'timepending': pl.Int64,
            'nonce': pl.Int64,
            'gas': pl.Int64,
            'gasprice': pl.Float64,
            'value': pl.Float64,
            'toaddress': pl.String,
            'fromaddress': pl.String,
            'input': pl.String,
            'network': pl.String,
            'type': pl.Int64,
            'maxpriorityfeepergas': pl.Float64,
            'maxfeepergas': pl.Float64,
            'basefeepergas': pl.Float64,
            'dropreason': pl.String,
            'rejectionreason': pl.String,
            'stuck': pl.Boolean,
            'gasused': pl.Int64,
            'detect_date': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        url = url_template.format(
            year=chunk.year,  # type: ignore
            month=chunk.month,  # type: ignore
            day=chunk.day,  # type: ignore
            hour=chunk.hour,  # type: ignore
        )
        polars_kwargs = {'separator': '\t', 'schema': self.get_schema()}
        return absorb.ops.download_csv_gz_to_dataframe(
            url=url, polars_kwargs=polars_kwargs
        )

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        return (
            datetime.datetime(year=2019, month=11, day=1, hour=0),
            datetime.datetime(year=2025, month=3, day=1, hour=0),
        )



from __future__ import annotations

import typing

import absorb
from .. import cli_outputs
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def add_command(args: Namespace) -> dict[str, Any]:
    import json
    import rich

    # parse inputs
    track_datasets = cli_parsing._parse_datasets(args)

    # add untracked collected
    if args.collected:
        track_datasets += [
            absorb.Table.instantiate(table_dict)
            for table_dict in absorb.ops.get_untracked_collected_tables()
        ]

    # filter already collected
    tracked_tables = absorb.ops.get_tracked_tables()
    tracked = [json.dumps(table, sort_keys=True) for table in tracked_tables]
    already_tracked = {}
    not_tracked = {}
    for ds in track_datasets:
        ds_hash = json.dumps(ds.create_table_dict(), sort_keys=True)
        if ds_hash in tracked:
            already_tracked[ds_hash] = ds
        else:
            not_tracked[ds_hash] = ds
    track_datasets = list(not_tracked.values())

    # check for invalid datasets
    sources = set(td.source for td in track_datasets)
    source_datasets = {
        source: absorb.ops.get_source_table_classes(source)
        for source in sources
    }
    for track_dataset in track_datasets:
        if type(track_dataset) not in source_datasets[track_dataset.source]:
            raise Exception('invalid dataset:')

    # print dataset summary
    if len(already_tracked) > 0:
        cli_outputs._print_title('Already tracking')
        for dataset in already_tracked.values():
            cli_outputs._print_dataset_bullet(dataset)
        print()
    cli_outputs._print_title('Now tracking')
    if len(track_datasets) == 0:
        print('[no new datasets specified]')
    else:
        for dataset in track_datasets:
            cli_outputs._print_dataset_bullet(dataset)
        print()
        rich.print(
            'to proceed with data collection, use [white bold]absorb collect[/white bold]'
        )

    # setup directories
    for table in track_datasets:
        table.setup_table_dir()

    # start tracking tables
    absorb.ops.start_tracking_tables(track_datasets)

    # check for missing packages or credentials
    warnings = []
    for table in track_datasets:
        name = table.full_name()

        missing_packages = table.get_missing_packages()
        for package in missing_packages:
            warnings.append(
                f'[red]missing package[/red]: [white]{package}[/white] for [yellow]{name}[/yellow]'
            )

        missing_credentials = table.get_missing_credentials()
        for credential in missing_credentials:
            warnings.append(
                f'[red]missing credentials[/red]: [white]{credential}[/white] for [yellow]{name}[/yellow]'
            )
    if len(warnings) > 0:
        print()
        rich.print('[red]Warnings:[/red]')
        for warning in warnings:
            rich.print('- ' + warning)

    return {}



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import datetime
    import polars as pl


url_template = 'https://kalshi-public-docs.s3.amazonaws.com/reporting/market_data_{year}-{month:02}-{day:02}.json'
path_template = '/Users/stormslivkoff/data/kalshi/raw_archive/market_data_{year}-{month:02}-{day:02}.json'


class Metrics(absorb.Table):
    source = 'kalshi'
    description = 'Daily summary data for each Kalshi market'
    url = 'https://kalshi.com/'
    write_range = 'append_only'
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'date': pl.String,
            'ticker_name': pl.String,
            'old_ticker_name': pl.String,
            'report_ticker': pl.String,
            'payout_type': pl.String,
            'open_interest': pl.Float64,
            'daily_volume': pl.Int64,
            'block_volume': pl.Int64,
            'high': pl.Int64,
            'low': pl.Int64,
            'status': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import requests
        import polars as pl

        date: datetime.datetime = chunk  # type: ignore
        url = get_date_url(date)
        response = requests.get(url, stream=True)
        if response.status_code == 404:
            return None
        response.raise_for_status()

        df = pl.DataFrame(response.json())
        if 'old_ticker_name' not in df.columns:
            df = df.insert_column(
                2, pl.lit(None, dtype=pl.String).alias('old_ticker_name')
            )

        return df

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        first = datetime.datetime(year=2021, month=6, day=30)
        last = _find_last()
        return (first, last)


class Metadata(absorb.Table):
    source = 'kalshi'
    description = 'Metadata for each Kalshi market'
    url = 'https://kalshi.com/'
    cadence = None
    write_range = 'overwrite_all'
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'series_ticker': pl.String,
            'series_title': pl.String,
            'total_series_volume': pl.Int64,
            'total_volume': pl.Int64,
            'event_ticker': pl.String,
            'event_subtitle': pl.String,
            'event_title': pl.String,
            'category': pl.String,
            'total_market_count': pl.Int64,
            # 'product_metadata': pl.Struct,
            # 'markets': pl.List,
            'is_trending': pl.Boolean,
            'is_new': pl.Boolean,
            'is_closing': pl.Boolean,
            'is_price_delta': pl.Boolean,
            'search_score': pl.Int64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import requests
        import time
        import polars as pl

        base_url = 'https://api.elections.kalshi.com/v1/search/series?order_by=newest&page_size=100'

        cursor = None
        cursor_results: list[typing.Any] = []
        while True:
            if cursor is not None:
                url = base_url + '&cursor=' + cursor
            else:
                url = base_url
            time.sleep(0.25)
            print('getting kalshi metadata page', len(cursor_results) + 1)
            response = requests.get(url)
            data = response.json()
            if response.status_code == 200:
                cursor_results.append(data)
                cursor = data.get('next_cursor')
                if cursor is None:
                    break
                if len(cursor_results) * 100 > 2 * data['total_results_count']:
                    break
            else:
                print('status code', response.status_code)
                break

        return (
            pl.DataFrame(
                item
                for result in cursor_results
                for item in result['current_page']
            )
            .select(self.get_schema().keys())
            .unique('series_ticker')
        )

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        now = datetime.datetime.now(datetime.timezone.utc).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        return (None, now)


def get_date_url(date: datetime.datetime) -> str:
    return url_template.format(year=date.year, month=date.month, day=date.day)


def get_date_path(date: datetime.datetime) -> str:
    return path_template.format(year=date.year, month=date.month, day=date.day)


def _find_last() -> datetime.datetime:
    import datetime

    current = datetime.datetime.now()
    current = datetime.datetime(
        year=current.year, month=current.month, day=current.day
    )
    while current > datetime.datetime(year=2021, month=6, day=28):
        if absorb.ops.does_remote_file_exist(get_date_url(current)):
            return current
        current = current - datetime.timedelta(days=1)
    raise Exception()



"""
instructions:
1. obtain json manifest from https://export.verifieralliance.org/manifest.json
2. download the files in the manifest
"""

from __future__ import annotations

import functools
import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


class VeraChunkedDataset(absorb.Table):
    source = 'vera'
    vera_filetype: str
    url = 'https://verifieralliance.org/'
    index_type = 'number_range'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        return {}

    def get_available_range(self) -> absorb.Coverage:
        return get_current_files(self.vera_filetype)

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        url = 'https://export.verifieralliance.org/' + chunk  # type: ignore
        return absorb.ops.download_parquet_to_dataframe(url=url)


class Code(VeraChunkedDataset):
    description = 'Verifier Alliance code dataset'
    write_range = 'overwrite_all'
    vera_filetype = 'code'


class Contracts(VeraChunkedDataset):
    description = 'Verifier Alliance contracts dataset'
    write_range = 'overwrite_all'
    vera_filetype = 'contracts'


class ContractDeployments(VeraChunkedDataset):
    description = 'Verifier Alliance contract deployments dataset'
    write_range = 'overwrite_all'
    vera_filetype = 'contract_deployments'


class CompiledContracts(VeraChunkedDataset):
    description = 'Verifier Alliance compiled contracts dataset'
    write_range = 'overwrite_all'
    vera_filetype = 'compiled_contracts'


class CompiledContractsSources(VeraChunkedDataset):
    description = 'Verifier Alliance compiled contract sourcecode dataset'
    write_range = 'overwrite_all'
    vera_filetype = 'compiled_contracts_sources'


class Sources(VeraChunkedDataset):
    description = 'Verifier Alliance sourcecode dataset'
    write_range = 'overwrite_all'
    vera_filetype = 'sources'


class VerifiedContracts(VeraChunkedDataset):
    description = 'Verifier Alliance verified contracts dataset'
    write_range = 'overwrite_all'
    vera_filetype = 'verified_contracts'


def get_tables() -> list[type[absorb.Table]]:
    return [
        Code,
        Contracts,
        ContractDeployments,
        CompiledContracts,
        CompiledContractsSources,
        Sources,
        VerifiedContracts,
    ]


@functools.lru_cache
def get_current_manifest() -> dict[str, typing.Any]:
    import requests

    url = 'https://export.verifieralliance.org/manifest.json'
    response = requests.get(url)
    response.raise_for_status()
    manifest: dict[str, typing.Any] = response.json()
    return manifest


def get_current_files(filetype: str) -> list[str]:
    manifest = get_current_manifest()
    if filetype in [
        'code',
        'contracts',
        'contract_deployments',
        'compiled_contracts',
        'compiled_contracts_sources',
        'sources',
        'verified_contracts',
    ]:
        files = manifest['files'][filetype]
        if not isinstance(files, list) or not all(
            isinstance(item, str) for item in files
        ):
            raise Exception()
        return files
    else:
        raise Exception('invalid filetype')



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl

import absorb
from . import common


# class DexVolumes(absorb.Table):
#     source = 'defillama'
#     write_range = 'overwrite_all'
#     parameter_types = {
#         'protocols': typing.Union[list[str], None],
#         'chains': typing.Union[list[str], None],
#         'top_n_protocols': typing.Union[int, None],
#         'top_n_chains': typing.Union[int, None],
#     }
#     default_parameters = {'protocols': None, 'chains': None}
#     name_template = {
#         'default': 'dex_volumes',
#         'custom': {
#             'protocols': 'dex_volumes_{protocols}',
#             'chains': 'dex_volumes_{chains}',
#             'top_n_protocols': 'dex_volumes_top_{top_n_protocols}_protocols',
#             'top_n_chains': 'dex_volumes_top_{top_n_chains}_chains',
#         },
#     }
#     index_type = 'date_range'

#     def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType] ]:
#         return {
#             'timestamp': pl.Datetime('ms'),
#             'chain': pl.String,
#             'protocol': pl.String,
#             'volume_usd': pl.Float64,
#         }

#     def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
#         import polars as pl

#         chains = self.parameters['chains']
#         protocols = self.parameters['protocols']

#         dfs = []
#         if protocols is not None:
#             print('collecting', len(protocols), 'protocols')
#             for p, protocol in enumerate(protocols, start=1):
#                 print(
#                     '[' + str(p) + ' / ' + str(len(protocols)) + ']', protocol
#                 )
#                 try:
#                     df = get_historical_dex_volume_per_chain_of_protocol(
#                         protocol
#                     )
#                     dfs.append(df)
#                 except Exception:
#                     print('could not collect', protocol)
#         else:
#             if chains is None:
#                 chains = _get_dex_chains()
#             print('collecting', len(chains), 'chains')
#             for c, chain in enumerate(chains, start=1):
#                 print('[' + str(c) + ' / ' + str(len(chains)) + ']', chain)
#                 try:
#                     df = get_historical_dex_volume_per_protocol_of_chain(chain)
#                     dfs.append(df)
#                 except Exception:
#                     print('could not collect', chain)
#         return pl.concat(dfs)


# class OptionsVolumes(absorb.Table):
#     source = 'defillama'
#     write_range = 'overwrite_all'
#     parameter_types = {
#         'protocols': typing.Union[list[str], None],
#         'chains': typing.Union[list[str], None],
#     }
#     default_parameters = {'protocols': None, 'chains': None}
#     index_type = 'date_range'

#     def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
#         return {
#             'timestamp': pl.Datetime('ms'),
#             'chain': pl.String,
#             'protocol': pl.String,
#             'volume_usd': pl.Float64,
#         }

#     def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
#         import polars as pl

#         chains = self.parameters['chains']
#         protocols = self.parameters['protocols']

#         dfs = []
#         if protocols is not None:
#             print('collecting', len(protocols), 'protocols')
#             for p, protocol in enumerate(protocols, start=1):
#                 print(
#                     '[' + str(p) + ' / ' + str(len(protocols)) + ']', protocol
#                 )
#                 try:
#                     df = get_historical_options_volume_per_chain_of_protocol(
#                         protocol
#                     )
#                     dfs.append(df)
#                 except Exception:
#                     print('could not collect', protocol)
#         else:
#             if chains is None:
#                 chains = _get_options_chains()
#             print('collecting', len(chains), 'chains')
#             for c, chain in enumerate(chains, start=1):
#                 print('[' + str(c) + ' / ' + str(len(chains)) + ']', chain)
#                 try:
#                     df = get_historical_options_volume_per_protocol_of_chain(
#                         chain
#                     )
#                     dfs.append(df)
#                 except Exception:
#                     print('could not collect', chain)
#         return pl.concat(dfs)


class ProtocolDexVolumes(absorb.Table):
    source = 'defillama'
    description = 'Volume of each dex protocol on each chain in USD over time'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'protocols': (list, type(None))}
    default_parameters = {'protocols': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'chain': pl.String,
            'protocol': pl.String,
            'volume_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        protocols = self.parameters['protocols']
        if protocols is None:
            protocols = _get_dex_protocols()
        dfs = []
        print('collecting', len(protocols), 'protocols')
        for p, protocol in enumerate(protocols, start=1):
            print('[' + str(p) + ' / ' + str(len(protocols)) + ']', protocol)
            try:
                df = get_historical_dex_volume_per_chain_of_protocol(protocol)
                dfs.append(df)
            except Exception:
                print('could not collect', protocol)
        return pl.concat(dfs)


class ChainDexVolumes(absorb.Table):
    source = 'defillama'
    description = 'Volume of each dex protocol on each chain in USD over time'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'chains': (list, type(None))}
    default_parameters = {'chains': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'chain': pl.String,
            'protocol': pl.String,
            'volume_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        chains = self.parameters['chains']
        if chains is None:
            chains = _get_dex_chains()
        dfs = []
        print('collecting', len(chains), 'chains')
        for c, chain in enumerate(chains, start=1):
            print('[' + str(c) + ' / ' + str(len(chains)) + ']', chain)
            try:
                df = get_historical_dex_volume_per_protocol_of_chain(chain)
                dfs.append(df)
            except Exception:
                print('could not collect', chain)
        return pl.concat(dfs)


class ProtocolOptionsVolumes(absorb.Table):
    source = 'defillama'
    description = (
        'Volume of each options protocol on each chain in USD over time'
    )
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'protocols': (list, type(None))}
    default_parameters = {'protocols': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'chain': pl.String,
            'protocol': pl.String,
            'volume_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        protocols = self.parameters['protocols']
        if protocols is None:
            protocols = _get_options_protocols()
        dfs = []
        print('collecting', len(protocols), 'protocols')
        for p, protocol in enumerate(protocols, start=1):
            print('[' + str(p) + ' / ' + str(len(protocols)) + ']', protocol)
            df = get_historical_options_volume_per_chain_of_protocol(protocol)
            dfs.append(df)
        return pl.concat(dfs)


class ChainOptionsVolumes(absorb.Table):
    source = 'defillama'
    description = (
        'Volume of each options protocol on each chain in USD over time'
    )
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'chains': (list, type(None))}
    default_parameters = {'chains': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'chain': pl.String,
            'protocol': pl.String,
            'volume_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        chains = self.parameters['chains']
        if chains is None:
            chains = _get_options_chains()
        dfs = []
        print('collecting', len(chains), 'chains')
        for c, chain in enumerate(chains, start=1):
            print('[' + str(c) + ' / ' + str(len(chains)) + ']', chain)
            df = get_historical_options_volume_per_protocol_of_chain(chain)
            dfs.append(df)
        return pl.concat(dfs)


#
# # dex volumes
#


def _get_dex_chains() -> list[str]:
    return (
        get_current_dex_volume_per_protocol()['chains']
        .list.explode()
        .unique()
        .sort()
        .to_list()
    )


def _get_dex_protocols() -> list[str]:
    return (
        get_current_dex_volume_per_protocol()['protocol']
        .unique()
        .sort()
        .to_list()
    )


def _get_options_chains() -> list[str]:
    return (
        get_current_options_volume_per_protocol()['chains']
        .list.explode()
        .unique()
        .sort()
        .to_list()
    )


def _get_options_protocols() -> list[str]:
    return (
        get_current_options_volume_per_protocol()['protocol']
        .unique()
        .sort()
        .to_list()
    )


def get_current_dex_volume_per_protocol(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('historical_dex_volume')

    rows = [
        [
            protocol['name'],
            protocol['displayName'],
            protocol['slug'],
            protocol['category'],
            protocol['chains'],
            protocol.get('total24h'),
            protocol.get('total48hto24h'),
            protocol.get('total7d'),
            protocol.get('total14dto7d'),
            protocol.get('total30d'),
            protocol.get('total1y'),
            protocol.get('totalAllTime'),
        ]
        for protocol in data['protocols']
    ]
    schema = {
        'protocol': pl.String,
        'displayName': pl.String,
        'slug': pl.String,
        'category': pl.String,
        'chains': None,
        'volume_24h_usd': pl.Float64,
        'volume_48h_to_24h_usd': pl.Float64,
        'volume_7d_usd': pl.Float64,
        'volume_14d_to_7d_usd': pl.Float64,
        'volume_30d_usd': pl.Float64,
        'volume_1y_usd': pl.Float64,
        'volume_all_time_usd': pl.Float64,
    }
    return pl.DataFrame(rows, schema=schema, orient='row')


def get_current_dex_volume_per_chain_per_protocol(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('historical_dex_volume')

    rows = []
    for protocol in data['protocols']:
        if protocol.get('breakdown24h') is None:
            continue
        for chain in protocol['breakdown24h'].keys():
            row = [
                chain,
                protocol['name'],
                protocol['displayName'],
                protocol['slug'],
                protocol['category'],
                protocol['protocolType'],
                sum(protocol['breakdown24h'][chain].values()),
                sum(protocol['breakdown30d'][chain].values()),
            ]
            rows.append(row)

    schema = {
        'chain': pl.String,
        'protocol': pl.String,
        'displayName': pl.String,
        'slug': pl.String,
        'category': pl.String,
        'protocolType': pl.String,
        'volume_24h_usd': pl.Float64,
        'volume_30d_usd': pl.Float64,
    }
    return pl.DataFrame(rows, schema=schema, orient='row')


def get_historical_dex_volume(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('historical_dex_volume')
    return pl.DataFrame(
        data['totalDataChart'], schema=['timestamp', 'volume_usd'], orient='row'
    ).with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))


def get_historical_dex_volume_per_protocol(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('historical_dex_volume')
    rows = [
        [timestamp, protocol, value]
        for timestamp, datum in data['totalDataChartBreakdown']
        for protocol, value in datum.items()
    ]
    schema = ['timestamp', 'protocol', 'volume_usd']
    return pl.DataFrame(rows, orient='row', schema=schema).with_columns(
        (pl.col.timestamp * 1000).cast(pl.Datetime('ms'))
    )


def get_historical_dex_volume_of_protocol(
    protocol: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch(
            'historical_dex_volume_of_protocol', {'protocol': protocol}
        )
    return pl.DataFrame(
        data['totalDataChart'],
        schema={'timestamp': pl.Int64, 'volume_usd': pl.Float64},
        orient='row',
    ).select(
        (pl.col.timestamp * 1000).cast(pl.Datetime('ms')),
        pl.lit(data['name']).alias('protocol'),
        'volume_usd',
    )


def get_historical_dex_volume_per_chain_of_protocol(
    protocol: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch(
            'historical_dex_volume_of_protocol', {'protocol': protocol}
        )
    rows = [
        [timestamp, chain, data['name'], sum(value.values())]
        for timestamp, datum in data['totalDataChartBreakdown']
        for chain, value in datum.items()
    ]
    schema = ['timestamp', 'chain', 'protocol', 'volume_usd']
    return pl.DataFrame(rows, orient='row', schema=schema).with_columns(
        (pl.col.timestamp * 1000).cast(pl.Datetime('ms'))
    )


def get_historical_dex_volume_of_chain(
    chain: str,
    *,
    data: pl.DataFrame | None = None,
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('historical_dex_volume_of_chain', {'chain': chain})

    return pl.DataFrame(
        data['totalDataChart'], schema=['timestamp', 'volume_usd'], orient='row'
    ).select(
        (pl.col.timestamp * 1000).cast(pl.Datetime('ms')),
        pl.lit(chain).alias('chain'),
        'volume_usd',
    )


def get_historical_dex_volume_per_protocol_of_chain(
    chain: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch('historical_dex_volume_of_chain', {'chain': chain})
    rows = [
        [timestamp, chain, protocol, value]
        for timestamp, datum in data['totalDataChartBreakdown']
        for protocol, value in datum.items()
    ]
    schema = ['timestamp', 'chain', 'protocol', 'volume_usd']
    return pl.DataFrame(rows, orient='row', schema=schema).with_columns(
        (pl.col.timestamp * 1000).cast(pl.Datetime('ms'))
    )


#
# # options volumes
#


def get_current_options_volume_per_protocol(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    if data is None:
        data = common._fetch('historical_options_volume')
    return get_current_dex_volume_per_protocol(data=data)


def get_current_options_volume_per_chain_per_protocol(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    if data is None:
        data = common._fetch('historical_options_volume')
    return get_current_dex_volume_per_chain_per_protocol(data=data)


def get_historical_options_volume(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    if data is None:
        data = common._fetch('historical_options_volume')
    return get_historical_dex_volume(data=data)


def get_historical_options_volume_per_protocol(
    *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    if data is None:
        data = common._fetch('historical_options_volume')
    return get_historical_dex_volume_per_protocol(data=data)


def get_historical_options_volume_of_protocol(
    protocol: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    if data is None:
        data = common._fetch(
            'historical_options_volume_of_protocol', {'protocol': protocol}
        )
    return get_historical_dex_volume_of_protocol(protocol=protocol, data=data)


def get_historical_options_volume_per_chain_of_protocol(
    protocol: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    if data is None:
        data = common._fetch(
            'historical_options_volume_of_protocol', {'protocol': protocol}
        )
    return get_historical_dex_volume_per_chain_of_protocol(
        protocol=protocol, data=data
    )


def get_historical_options_volume_of_chain(
    chain: str,
    *,
    data: pl.DataFrame | None = None,
) -> pl.DataFrame:
    if data is None:
        data = common._fetch(
            'historical_options_volume_of_chain', {'chain': chain}
        )
    return get_historical_dex_volume_of_chain(chain=chain, data=data)


def get_historical_options_volume_per_protocol_of_chain(
    chain: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    if data is None:
        data = common._fetch(
            'historical_options_volume_of_chain', {'chain': chain}
        )
    return get_historical_dex_volume_per_protocol_of_chain(
        chain=chain, data=data
    )



# https://docs.growthepie.xyz/api

from __future__ import annotations

import absorb

import typing

if typing.TYPE_CHECKING:
    import polars as pl


root = 'https://l2beat.com/api/'

endpoints = {
    'summary': root + 'scaling/summary?range=max',
    'value': root + 'scaling/tvs?range=max',
    'project_tvs': root + 'scaling/tvs/{project}?range=max',
    'activity': root + 'scaling/activity?range=max',
    'project_activity': root + 'scaling/activity/{project}?range=max',
}


class Metrics(absorb.Table):
    source = 'l2beat'
    description = 'On-chain metrics for Ethereum and its rollups'
    url = 'https://l2beat.com/'
    write_range = 'overwrite_all'
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime(time_unit='ms', time_zone=None),
            'n_transactions': pl.Int64,
            'n_user_operations': pl.Int64,
            'native_tvs': pl.Float64,
            'canonical_tvs': pl.Float64,
            'external_tvs': pl.Float64,
            'total_tvs': pl.Float64,
            'chain': pl.String,
            'layer': pl.String,
            'category': pl.String,
            'da': pl.String,
            'stack': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_all_data()

    def get_available_range(self) -> absorb.Coverage:
        import datetime
        import requests

        first = datetime.datetime(year=2019, month=11, day=15)
        response = requests.get(endpoints['value'])
        data = response.json()
        last_timestamp = data['data']['chart']['data'][-1][0]
        last = datetime.datetime.fromtimestamp(last_timestamp)
        return (first, last)


def get_projects() -> pl.DataFrame:
    import requests
    import polars as pl

    response = requests.get(endpoints['summary'])
    data = response.json()
    return pl.DataFrame(list(data['projects'].values()), orient='row')


def get_project_activity(project: str) -> pl.DataFrame:
    import requests
    import polars as pl

    response = requests.get(
        endpoints['project_activity'].format(project=project)
    )
    data = response.json()
    if not data['success']:
        raise Exception(data['error'])
    return (
        pl.DataFrame(
            data['data']['chart']['data'],
            schema=data['data']['chart']['types'],
            orient='row',
        )
        .with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))
        .rename({'count': 'n_transactions', 'uopsCount': 'n_user_operations'})
    )


def get_project_tvs(project: str) -> pl.DataFrame:
    import requests
    import polars as pl

    response = requests.get(endpoints['project_tvs'].format(project=project))
    data = response.json()
    if not data['success']:
        raise Exception(data['error'])
    return pl.DataFrame(
        data['data']['chart']['data'],
        schema=data['data']['chart']['types'],
        orient='row',
    ).select(
        timestamp=(pl.col.timestamp * 1000).cast(pl.Datetime('ms')),
        native_tvs=pl.col.native.cast(pl.Float64),
        canonical_tvs=pl.col.canonical.cast(pl.Float64),
        external_tvs=pl.col.external.cast(pl.Float64),
        total_tvs=(pl.col.native + pl.col.canonical + pl.col.external).cast(
            pl.Float64
        ),
    )


def get_all_data(*, projects: pl.DataFrame | None = None) -> pl.DataFrame:
    import polars as pl
    import time

    if projects is None:
        projects = get_projects()

    dfs = []
    for project in projects.to_dicts():
        print('getting', project['slug'])
        try:
            time.sleep(20)
            activity = get_project_activity(project['slug'])
            time.sleep(20)
            tvs = get_project_tvs(project['slug'])
        except Exception as e:
            print('skipping ' + project['slug'] + ' because ' + str(e.args[0]))
            continue
        providers = project.get('providers')
        if providers is None:
            providers = []
        provider = ', '.join(providers)
        df = (
            activity.join(tvs, on='timestamp', how='full', coalesce=True)
            .with_columns(
                chain=pl.lit(project['name']),
                layer=pl.lit(project['type']),
                category=pl.lit(project['category']),
                da=pl.lit(project['hostChain']),
                stack=pl.lit(provider),
            )
            .sort('timestamp')
        )
        dfs.append(df)

    return pl.concat(dfs)



from __future__ import annotations

import typing
import absorb
from . import table_names

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')
    import polars as pl


class TablePaths(table_names.TableNames):
    def get_table_dir(self, warn: bool = True) -> str:
        return absorb.ops.paths.get_table_dir(
            source=self.source, table=self.name(), warn=warn
        )

    def get_glob(self, warn: bool = True) -> str:
        return self.get_file_path(glob=True, warn=warn)

    def get_file_path(
        self,
        chunk: absorb.Chunk | None = None,
        glob: bool = False,
        warn: bool = True,
        df: pl.DataFrame | None = None,
    ) -> str:
        # special case the chunk if write_range=overwrite_all
        if self.write_range == 'overwrite_all':
            if glob:
                chunk = None
            else:
                if df is not None:
                    if 'timestamp' in df.columns:
                        chunk = df['timestamp'].max()
                    else:
                        chunk = 'all'
                else:
                    raise Exception('must specify range')

        # get file path
        return absorb.ops.paths.get_table_filepath(
            chunk=chunk,
            index_type=self.index_type,
            filename_template=self.filename_template,
            table=self.name(),
            source=self.source,
            parameters=self.parameters,
            glob=glob,
            warn=warn,
        )

    def get_file_paths(
        self, chunks: absorb.Coverage, warn: bool = True
    ) -> list[str]:
        return absorb.ops.paths.get_table_filepaths(
            chunks=chunks,
            index_type=self.index_type,
            filename_template=self.filename_template,
            table=self.name(),
            source=self.source,
            parameters=self.parameters,
            warn=warn,
        )

    def parse_file_path(self, path: str) -> dict[str, typing.Any]:
        if self.write_range == 'overwrite_all':
            index_type = None
        else:
            index_type = self.index_type
        return absorb.ops.paths.parse_file_path(
            path=path,
            filename_template=self.filename_template,
            index_type=index_type,
        )

    def setup_table_dir(self) -> None:
        import json
        import os

        # create directory
        table_dir = self.get_table_dir()
        os.makedirs(table_dir, exist_ok=True)

        # set up metadata
        metadata = self.create_table_dict()
        metadata_path = absorb.ops.get_table_metadata_path(
            self.name(), source=self.source
        )
        if os.path.isfile(metadata_path):
            with open(metadata_path, 'r') as f:
                existing_metadata = json.load(f)
            if json.dumps(existing_metadata, sort_keys=True) != json.dumps(
                metadata, sort_keys=True
            ):
                raise ValueError(
                    'metadata for table does not match existing metadata in directory'
                )
        else:
            with open(metadata_path, 'w') as f:
                json.dump(metadata, f)
            if absorb.ops.get_config()['use_git']:
                absorb.ops.git_add_and_commit_file(
                    metadata_path,
                    repo_root=absorb.ops.get_absorb_root(),
                    message='Collect new table metadata: ' + self.full_name(),
                )

    def create_table_dict(self) -> absorb.TableDict:
        """
        Create metadata for the table file.
        """
        return {
            'source_name': self.source,
            'table_version': self.version,
            'table_name': self.name(),
            'table_class': type(self).__module__ + '.' + type(self).__name__,
            'parameters': self.parameters,
        }



from __future__ import annotations

import typing
import absorb

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')

    import polars as pl
    import tooltime


class TableBase:
    version: str = '0.1.0'
    source: str
    description: str | None = None
    url: str | None = None

    #
    # # structure
    #

    write_range: typing.Literal[
        'append_only', 'overwrite_all', 'overwrite_chunks'
    ]
    index_type: absorb.IndexType | None = None

    # for ongoing datasets, time to wait before checking if new data is posted
    # can be a str duration like '1h' or a float multiple of index type
    update_latency: tooltime.Timelength | float | None = None

    chunk_datatype: typing.Literal['dataframe', 'files'] = 'dataframe'

    # dependencies
    required_packages: list[str] = []
    required_credentials: list[str] = []

    #
    # # parameters
    #

    parameter_types: dict[str, type | tuple[type, ...]] = {}
    default_parameters: dict[str, absorb.JSONValue] = {}
    parameters: dict[str, typing.Any]

    #
    # # naming
    #

    # use first available template that has all parameters filled
    name_template: str | list[str] = '{class_name}'
    filename_template = '{source}__{table}__{chunk}.parquet'

    #
    # # methods
    #

    def get_schema(self) -> dict[str, type[pl.DataType] | pl.DataType]:
        raise NotImplementedError()

    @classmethod
    def get_missing_packages(cls) -> list[str]:
        return [
            package
            for package in cls.required_packages
            if not absorb.ops.is_package_installed(package)
        ]

    @classmethod
    def get_missing_credentials(self) -> list[str]:
        import os

        missing = []
        for credential in self.required_credentials:
            value = os.environ.get(credential)
            if value is None or value == '':
                missing.append(credential)
        return missing



from __future__ import annotations

import typing

import absorb
from .. import cli_outputs
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def setup_command(args: Namespace) -> dict[str, Any]:
    import toolstr
    import os

    # get list of input tables
    if len(args.dataset) > 0:
        datasets = cli_parsing._parse_datasets(args)
    else:
        datasets = [
            absorb.Table.instantiate(table_dict)
            for table_dict in absorb.ops.get_tracked_tables()
        ]

    # regenerate table metadata
    if args.regenerate_metadata:
        if len(datasets) == 1:
            word = 'dataset'
        else:
            word = 'datasets'
        print('generating metadata for', len(datasets), word)
        for dataset in datasets:
            instance = absorb.Table.instantiate(dataset)
            instance.setup_table_dir()

    # regenerate config
    if args.regenerate_config:
        old_config = absorb.ops.get_config()
        new_config = absorb.ops.get_default_config()
        for tracked_table in old_config['tracked_tables']:
            table = absorb.Table.instantiate(tracked_table)
            table_dict = table.create_table_dict()
            new_config['tracked_tables'].append(table_dict)
        absorb.ops.write_config(new_config)

    # change git settings
    if args.disable_git:
        absorb.ops.disable_git_tracking()
        config = absorb.ops.get_config()
    if args.enable_git:
        absorb.ops.enable_git_tracking()
        config = absorb.ops.get_config()

    # setup git tracking
    config = absorb.ops.get_config()
    if config['use_git']:
        absorb.ops.setup_git(track_datasets=datasets)

    if args.set_default_rclone_remote:
        absorb.ops.set_default_rclone_remote(args.set_default_rclone_remote)
    if args.set_default_bucket:
        absorb.ops.set_default_bucket(args.set_default_bucket)
    if args.set_default_provider:
        absorb.ops.set_default_provider(args.set_default_provider)
    if args.set_default_path_prefix:
        absorb.ops.set_default_path_prefix(args.set_default_path_prefix)
    if args.clear_default_rclone_remote:
        absorb.ops.clear_default_rclone_remote()
    if args.clear_default_bucket:
        absorb.ops.clear_default_bucket()
    if args.clear_default_provider:
        absorb.ops.clear_default_provider()
    if args.clear_default_path_prefix:
        absorb.ops.clear_default_path_prefix()

    # print config
    toolstr.print_text_box(
        'Current config', style='green', text_style='bold white'
    )
    config = absorb.ops.get_config()
    for key, value in config.items():
        if key == 'tracked_tables':
            continue
        if isinstance(value, dict):
            absorb.ops.print_bullet(key=key, value='')
            for subkey, subvalue in value.items():
                absorb.ops.print_bullet(key=subkey, value=subvalue, indent=4)
        else:
            absorb.ops.print_bullet(key=key, value=str(value))
    print()
    names = [
        table['source_name'] + '.' + table['table_name']
        for table in config['tracked_tables']
    ]
    toolstr.print_header(
        'Tracked datasets (' + str(len(names)) + ')',
        style='green',
        text_style='bold white',
    )
    for n, name in enumerate(sorted(names)):
        if not args.verbose:
            if n == 5:
                print('...')
            if n > 4 and n != len(names) - 1:
                continue
        absorb.ops.print_bullet(key=None, value=name, number=n + 1)

    print()
    if 'ABSORB_ROOT' in os.environ:
        print('config stored at ABSORB_ROOT:', absorb.ops.get_absorb_root())
    else:
        print('using default config, set ABSORB_ROOT to a config')

    return {}



from __future__ import annotations

import typing

import absorb
from . import common

if typing.TYPE_CHECKING:
    import polars as pl


class ChainTvls(absorb.Table):
    source = 'defillama'
    description = 'TVL of each chain in USD over time'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'chains': (list, type(None))}
    default_parameters = {'chains': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'chain': pl.String,
            'tvl_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        chains = self.parameters['chains']
        if chains is None:
            chains = _get_tvl_chains()
        print('collecting', len(chains), 'chains')
        dfs = []
        for c, chain in enumerate(chains, start=1):
            print('[' + str(c) + ' / ' + str(len(chains)) + ']', chain)
            try:
                df = get_historical_tvl_of_chain(chain)
            except Exception:
                print('could not collect ' + chain)
                continue
            dfs.append(df)
        return pl.concat(dfs)


class ProtocolTvls(absorb.Table):
    source = 'defillama'
    description = 'TVL of each protocol on each chain in USD over time'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'protocols': (list, type(None))}
    default_parameters = {'protocols': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'chain': pl.String,
            'protocol': pl.String,
            'tvl_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        protocols = self.parameters['protocols']
        if protocols is None:
            protocols = _get_tvl_protocols()
        dfs = []
        print('collecting', len(protocols), 'protocols')
        for p, protocol in enumerate(protocols, start=1):
            print('[' + str(p) + ' / ' + str(len(protocols)) + ']', protocol)
            try:
                df = get_historical_tvl_per_chain_of_protocol(protocol)
                dfs.append(df)
            except Exception:
                print('could not collect', protocol)
                continue
        return pl.concat(dfs)


class ProtocolTvlsPerToken(absorb.Table):
    source = 'defillama'
    description = 'TVL of each token in each protocol in USD over time'
    url = 'https://defillama.com/'
    write_range = 'overwrite_all'
    parameter_types = {'protocols': (list, type(None))}
    default_parameters = {'protocols': None}
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('ms'),
            'protocol': pl.String,
            'symbol': pl.String,
            'supply': pl.Float64,
            'tvl_usd': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import polars as pl

        protocols = self.parameters['protocols']
        if protocols is None:
            protocols = _get_tvl_protocols()
        dfs = []
        print('collecting', len(protocols), 'protocols')
        for p, protocol in enumerate(protocols, start=1):
            print('[' + str(p) + ' / ' + str(len(protocols)) + ']', protocol)
            df = get_historical_tvl_per_token_of_protocol(protocol)
            dfs.append(df)
        return pl.concat(dfs)


def _get_tvl_chains() -> list[str]:
    return (
        get_current_project_tvls()['chains']
        .list.explode()
        .unique()
        .sort()
        .drop_nulls()
        .to_list()
    )


def _get_tvl_protocols() -> list[str]:
    return (
        get_current_project_tvls()['protocol']
        .unique()
        .sort()
        .drop_nulls()
        .to_list()
    )


def get_current_project_tvls() -> pl.DataFrame:
    import polars as pl

    data = common._fetch('current_tvls')

    return pl.DataFrame(
        data, orient='row', infer_schema_length=len(data), strict=False
    ).select(
        pl.col('name').alias('protocol'),
        'slug',
        pl.col.parentProtocol.str.strip_prefix('parent#').alias('parent'),
        'category',
        (pl.col.listedAt * 1000).cast(pl.Datetime('ms')).alias('list_date'),
        'symbol',
        'chain',
        'chains',
        'url',
        'github',
        pl.col.tvl.cast(pl.Float64).alias('tvl_usd'),
    )


def get_historical_tvl() -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_tvl')
    return pl.DataFrame(data, orient='row').select(
        timestamp=(pl.col.date * 1000).cast(pl.Datetime('ms')),
        tvl_usd=pl.col.tvl.cast(pl.Float64),
    )


def get_historical_tvl_of_chain(chain: str) -> pl.DataFrame:
    import polars as pl

    data = common._fetch('historical_tvl_of_chain', {'chain': chain})
    return pl.DataFrame(data).select(
        timestamp=(pl.col.date * 1000).cast(pl.Datetime('ms')),
        chain=pl.lit(chain),
        tvl_usd=pl.col.tvl.cast(pl.Float64),
    )


def get_historical_tvl_per_chain_of_protocol(
    protocol: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch(
            'historical_tvl_of_protocol', {'protocol': protocol}
        )
    rows = [
        [datum['date'], chain, protocol, float(datum['totalLiquidityUSD'])]
        for chain in data['chainTvls']
        for datum in data['chainTvls'][chain]['tvl']
    ]
    schema = {
        'timestamp': pl.Float64,
        'chain': pl.String,
        'protocol': pl.String,
        'tvl_usd': pl.Float64,
    }
    return pl.DataFrame(rows, schema=schema, orient='row').with_columns(
        (pl.col.timestamp * 1000).cast(pl.Datetime('ms'))
    )


def get_historical_tvl_per_token_of_protocol(
    protocol: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch(
            'historical_tvl_of_protocol', {'protocol': protocol}
        )

    rows = [
        [datum['date'], protocol, symbol, float(value)]
        for datum in data['tokens']
        for symbol, value in datum['tokens'].items()
    ]
    schema = ['timestamp', 'protocol', 'symbol', 'supply']
    tokens = (
        pl.DataFrame(rows, schema=schema, orient='row')
        .with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))
        .sort('timestamp', 'symbol')
    )

    rows = [
        [datum['date'], protocol, symbol, float(value)]
        for datum in data['tokensInUsd']
        for symbol, value in datum['tokens'].items()
    ]
    schema = ['timestamp', 'protocol', 'symbol', 'tvl_usd']
    tokensInUsd = (
        pl.DataFrame(rows, schema=schema, orient='row')
        .with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))
        .sort('timestamp', 'protocol', 'symbol')
    )

    return tokens.join(
        tokensInUsd, on=['timestamp', 'protocol', 'symbol'], how='inner'
    )


def get_historical_tvl_of_protocol(
    protocol: str, *, data: pl.DataFrame | None = None
) -> pl.DataFrame:
    import polars as pl

    if data is None:
        data = common._fetch(
            'historical_tvl_of_protocol', {'protocol': protocol}
        )
    rows = [
        [int(datum['date']), protocol, float(datum['totalLiquidityUSD'])]
        for datum in data['tvl']
    ]
    return pl.DataFrame(
        rows, schema=['timestamp', 'protocol', 'tvl_usd'], orient='row'
    ).with_columns((pl.col.timestamp * 1000).cast(pl.Datetime('ms')))



from __future__ import annotations

import typing

import absorb
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def collect_command(args: Namespace) -> dict[str, Any]:
    import toolstr

    # parse which datasets to collect
    if len(args.dataset) > 0:
        datasets = cli_parsing._parse_datasets(args)
    else:
        datasets = [
            absorb.Table.instantiate(table)
            for table in absorb.ops.get_tracked_tables()
        ]

    # print datasets to collect
    if len(datasets) == 0:
        import sys

        print('specify one or more datasets to collect')
        sys.exit(0)
    elif len(datasets) > 1:
        toolstr.print_text_box(
            'Collecting multiple datasets',
            style='green',
            text_style='bold white',
        )
        for d, dataset in enumerate(datasets):
            name = dataset.full_name()
            absorb.ops.print_bullet(key=name, value=None, number=d + 1)
        print()

    # collect each dataset
    first = True
    for dataset in datasets:
        if not first:
            print()

        # instantiate dataset
        if dataset.write_range == 'append_only':
            index_type = dataset.index_type
            if index_type is None:
                raise Exception(
                    'Cannot collect append_only dataset without index_type.'
                )
            data_ranges = cli_parsing._parse_ranges(
                args.range, index_type=index_type
            )
        else:
            data_ranges = None

        # collect dataset
        if args.setup_only:
            dataset.setup_table_dir()
        else:
            dataset.collect(
                data_range=data_ranges,
                dry=args.dry,
                overwrite=args.overwrite,
                verbose=args.verbose,
            )
        first = False

    return {}



from .buckets import *
from .catalog import *
from .config import *
from .coverage import *
from .formatting import *
from .git import *
from .io import *
from .names import *
from .networking import *
from .parsing import *
from .paths import *
from .queries import *
from .ranges import *
from .validation import *



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl

import absorb
from . import table_paths


class TableIO(table_paths.TablePaths):
    def scan(
        self,
        *,
        scan_kwargs: dict[str, typing.Any] | None = None,
    ) -> pl.LazyFrame:
        import polars as pl

        if scan_kwargs is None:
            scan_kwargs = {}
        try:
            return pl.scan_parquet(self.get_glob(), **scan_kwargs)
        except Exception as e:
            if e.args[0].startswith('expected at least 1 source'):
                raise Exception('no data to load for ' + str(self.full_name()))
            else:
                raise e

    def load(self, **kwargs: typing.Any) -> pl.DataFrame:
        """kwargs are the parameters of Table.scan()"""
        import polars as pl

        try:
            return self.scan(**kwargs).collect()
        except pl.exceptions.ComputeError as e:
            if e.args[0].startswith('expected at least 1 source'):
                raise Exception('no data to load for ' + str(self.full_name()))
            else:
                raise e



from __future__ import annotations

import typing

import absorb
from .. import cli_outputs

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def validate_command(args: Namespace) -> dict[str, Any]:
    import glob
    import json
    import os
    import sys
    import toolstr

    # collect errors
    instances = []
    errors = []
    metadatas = {}

    # check that config is valid
    config = absorb.ops.get_config()
    default_config = absorb.ops.get_default_config()
    assert config.keys() == default_config.keys(), (
        'config keys do not match default config keys'
    )
    for key, value in default_config.items():
        if isinstance(value, dict):
            assert config[key].keys() == value.keys(), (  # type: ignore
                f'config {key} subkeys do not match subkeys of '
                + ', '.join(value.keys())
            )

    # check that datasets_dir exists
    datasets_dir = absorb.ops.get_datasets_dir()
    if not os.path.isdir(datasets_dir):
        print('no datasets collected')
        sys.exit(0)

    # check each collected dataset
    for source in os.listdir(datasets_dir):
        tables_dir = absorb.ops.get_source_tables_dir(source)
        for table in os.listdir(tables_dir):
            # check that each dataset has a metadata file
            table_dir = os.path.join(tables_dir, table)
            metadata_path = absorb.ops.get_table_metadata_path(
                table, source=source
            )
            if not os.path.isfile(metadata_path):
                errors.append(f'{source}/{table} is missing a metadata file')

            # load metadata
            with open(metadata_path, 'r') as f:
                metadata = json.load(f)
            metadatas[table_dir] = metadata

            # validate that metadata is valid
            errors += [
                source + '.' + table + ' ' + error
                for error in absorb.ops.validate_table_dict(metadata)
            ]

            # instantiate table
            instance = absorb.Table.instantiate(metadata)
            instances.append(instance)

            # check that metadata matches table definition
            if instance.source != metadata['source_name']:
                errors.append(
                    source
                    + '.'
                    + table
                    + ' metadata source '
                    + metadata['source_name']
                    + ' does not match table source'
                    + instance.source
                )
            if instance.name() != metadata['table_name']:
                errors.append(
                    source
                    + '.'
                    + table
                    + ' metadata table name '
                    + metadata['table_name']
                    + ' does not match instance name '
                    + instance.name()
                )
            if instance.full_class_name() != metadata['table_class']:
                errors.append(
                    source
                    + '.'
                    + table
                    + ' metadata class name '
                    + str(metadata['table_class'])
                    + ' does not match instance class name '
                    + instance.full_class_name()
                )
            if instance.version != metadata['table_version']:
                errors.append(
                    source
                    + '.'
                    + table
                    + ' metadata table version '
                    + str(metadata['table_version'])
                    + ' does not match instance table version '
                    + str(instance.version)
                )

            # check that metadata matches filesystem location
            if source != metadata['source_name']:
                errors.append(
                    source
                    + '.'
                    + table
                    + ' metadata source '
                    + metadata['source_name']
                    + ' does not match filesystem location '
                    + table_dir
                )
            if table != metadata['table_name']:
                errors.append(
                    source
                    + '.'
                    + table
                    + ' metadata table name '
                    + metadata['table_name']
                    + ' does not match filesystem location '
                    + table_dir
                )

            # check there are no extra files beyond metadata and parquet files
            target_parquet_files = glob.glob(instance.get_glob())
            for filename in os.listdir(table_dir):
                if filename == os.path.basename(metadata_path):
                    continue
                path = os.path.join(table_dir, filename)
                if path not in target_parquet_files:
                    errors.append(
                        source
                        + '.'
                        + table
                        + ' has unexpected file '
                        + filename
                        + ' in directory '
                        + table_dir
                    )

    # validate tracked datasets
    config = absorb.ops.get_config()
    for metadata in config['tracked_tables']:
        # check that metadata is valid
        errors += [
            metadata.get('source_name', 'unknown_source')
            + '.'
            + metadata.get('table_name', 'unknown_table')
            + ' '
            + error
            for error in absorb.ops.validate_table_dict(metadata)
        ]

        # check that tracked metadata directory exists
        table_dir = absorb.ops.get_table_dir(metadata)
        if not os.path.isdir(table_dir):
            errors.append(
                metadata.get('source_name', 'unknown_source')
                + '.'
                + metadata.get('table_name', 'unknown_table')
                + ' directory does not exist: '
                + table_dir
            )
            continue

        # check that tracked metadata matches table folder metadata
        if json.dumps(metadata, sort_keys=True) != json.dumps(
            metadatas.get(table_dir, {}), sort_keys=True
        ):
            errors.append(
                metadata.get('source_name', 'unknown_source')
                + '.'
                + metadata.get('table_name', 'unknown_table')
                + ' tracked metadata does not match table folder metadata'
            )

    # print errors
    if len(errors) > 0:
        toolstr.print_text_box('Errors Found', style='red')
        for e, error in enumerate(errors):
            toolstr.print_bullet('[red]' + error + '[/red]', number=e + 1)
    else:
        toolstr.print('no errors found', style='green')

    return {}



"""functions for fetching data from https://data.binance.vision/"""

from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import datetime
    import polars as pl


CandleInterval = typing.Literal[
    '1s',
    '1m',
    '3m',
    '5m',
    '15m',
    '30m',
    '1h',
    '2h',
    '4h',
    '6h',
    '8h',
    '12h',
    '1d',
]


class SpotCandles(absorb.Table):
    source = 'binance'
    description = 'OHLCV candles for spot pairs at various time intervals'
    url = 'https://data.binance.vision/?prefix=data/spot/daily/klines/'
    write_range = 'append_only'
    index_type = 'day'
    parameter_types = {'pair': str, 'interval': str}
    default_parameters = {}
    name_template = 'spot_candles_{pair}_{interval}'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('us'),
            'pair': pl.String,
            'open': pl.Float64,
            'high': pl.Float64,
            'low': pl.Float64,
            'close': pl.Float64,
            'n_trades': pl.Int64,
            'base_volume': pl.Float64,
            'quote_volume': pl.Float64,
            'taker_buy_base_volume': pl.Float64,
            'taker_buy_quote_volume': pl.Float64,
        }

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        return (
            datetime.datetime(2025, 1, 1),
            datetime.datetime(2025, 5, 1),
        )

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_spot_candles(
            pair=self.parameters['pair'],
            timestamp=chunk,  # type: ignore
            interval=self.parameters['interval'],
            window='daily',
        )


class SpotTrades(absorb.Table):
    source = 'binance'
    description = 'Trades for a given spot pair'
    url = 'https://data.binance.vision/?prefix=data/spot/daily/trades/'
    write_range = 'append_only'
    index_type = 'day'
    parameter_types = {'pair': str}
    default_parameters = {}
    name_template = 'spot_trades_{pair}'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('us'),
            'pair': pl.String,
            'price': pl.Float64,
            'quantity_base': pl.Float64,
            'quantity_quote': pl.Float64,
            'buyer_is_maker': pl.Boolean,
            'best_price_match': pl.Boolean,
            'trade_id': pl.Int64,
        }

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        return (
            datetime.datetime(2025, 1, 1),
            datetime.datetime(2025, 5, 1),
        )

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_spot_trades(
            pair=self.parameters['pair'],
            timestamp=chunk,  # type: ignore
            window='daily',
        )


class SpotAggregateTrades(absorb.Table):
    source = 'binance'
    description = (
        'Trades aggregated by price for spot pairs on short time scales'
    )
    url = 'https://data.binance.vision/?prefix=data/spot/daily/aggTrades/'
    write_range = 'append_only'
    index_type = 'day'
    parameter_types = {'pair': str}
    default_parameters = {}
    name_template = 'spot_aggregate_trades_{pair}'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'timestamp': pl.Datetime('us'),
            'pair': pl.String,
            'price': pl.Float64,
            'quantity': pl.Float64,
            'buyer_is_maker': pl.Boolean,
            'best_price_match': pl.Boolean,
            'aggregate_trade_id': pl.Int64,
            'first_trade_id': pl.Int64,
            'last_trade_id': pl.Int64,
        }

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        return (
            datetime.datetime(2025, 1, 1),
            datetime.datetime(2025, 5, 1),
        )

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        return get_spot_aggregate_trades(
            pair=self.parameters['pair'],
            timestamp=chunk,  # type: ignore
            window='daily',
        )


def get_spot_url(
    *,
    pair: str,
    timestamp: datetime.datetime,
    datatype: typing.Literal['trades', 'aggTrades', 'klines'],
    window: typing.Literal['daily', 'monthly'],
    interval: CandleInterval | None = None,
) -> str:
    import datetime
    import os

    if window == 'daily':
        if timestamp != datetime.datetime(
            timestamp.year, timestamp.month, timestamp.day
        ):
            raise Exception('timestamp must be a specific day')
        date_str = timestamp.strftime('%Y-%m-%d')
    elif window == 'monthly':
        if timestamp != datetime.datetime(timestamp.year, timestamp.month, 1):
            raise Exception('timestamp must be a specific month')
        date_str = timestamp.strftime('%Y-%m')
    else:
        raise Exception('invalid interval, choose daily or monthly')

    if datatype == 'klines':
        if interval is None:
            raise Exception('must specify interval')
        template = 'spot/{window}/klines/{pair}/{interval}/{pair}-{interval}-{date_str}.zip'
    else:
        if interval is not None:
            raise Exception(
                'cannot specify interval for dataset, only specify window'
            )
        template = (
            'spot/{window}/{datatype}/{pair}/{pair}-{datatype}-{date_str}.zip'
        )

    root = 'https://data.binance.vision/data/'
    tail = template.format(
        datatype=datatype,
        pair=pair,
        window=window,
        interval=interval,
        date_str=date_str,
    )
    return root + tail


def get_spot_trades(
    pair: str,
    timestamp: datetime.datetime,
    window: typing.Literal['daily', 'monthly'] = 'daily',
) -> pl.DataFrame:
    import polars as pl

    url = get_spot_url(
        pair=pair,
        timestamp=timestamp,
        datatype='trades',
        window=window,
    )

    raw_schema: dict[str, pl.DataType | type[pl.DataType]] = {
        'trade_id': pl.Int64,
        'price': pl.Float64,
        'quantity_base': pl.Float64,
        'quantity_quote': pl.Float64,
        'timestamp': pl.Int64,
        'buyer_is_maker': pl.Boolean,
        'best_price_match': pl.Boolean,
    }

    columns: list[str | pl.Expr] = [
        'timestamp',
        pl.lit(pair).alias('pair'),
        'price',
        'quantity_base',
        'quantity_quote',
        'buyer_is_maker',
        'best_price_match',
        'trade_id',
    ]

    return _process(url=url, raw_schema=raw_schema, columns=columns)


def get_spot_aggregate_trades(
    pair: str,
    timestamp: datetime.datetime,
    window: typing.Literal['daily', 'monthly'] = 'daily',
) -> pl.DataFrame:
    import polars as pl

    url = get_spot_url(
        pair=pair,
        timestamp=timestamp,
        datatype='aggTrades',
        window=window,
    )

    raw_schema: dict[str, pl.DataType | type[pl.DataType]] = {
        'aggregate_trade_id': pl.Int64,
        'price': pl.Float64,
        'quantity': pl.Float64,
        'first_trade_id': pl.Int64,
        'last_trade_id': pl.Int64,
        'timestamp': pl.Int64,
        'buyer_is_maker': pl.Boolean,
        'best_price_match': pl.Boolean,
    }

    columns: list[str | pl.Expr] = [
        'timestamp',
        pl.lit(pair).alias('pair'),
        'price',
        'quantity',
        'buyer_is_maker',
        'best_price_match',
        'aggregate_trade_id',
        'first_trade_id',
        'last_trade_id',
    ]

    return _process(url=url, raw_schema=raw_schema, columns=columns)


def get_spot_candles(
    pair: str,
    timestamp: datetime.datetime,
    interval: CandleInterval,
    window: typing.Literal['daily', 'monthly'] = 'daily',
) -> pl.DataFrame:
    import polars as pl

    url = get_spot_url(
        pair=pair,
        timestamp=timestamp,
        datatype='klines',
        interval=interval,
        window=window,
    )

    raw_schema: dict[str, pl.DataType | type[pl.DataType]] = {
        'timestamp': pl.Int64,
        'open': pl.Float64,
        'high': pl.Float64,
        'low': pl.Float64,
        'close': pl.Float64,
        'base_volume': pl.Float64,
        'close_timestamp': pl.Int64,
        'quote_volume': pl.Float64,
        'n_trades': pl.Int64,
        'taker_buy_base_volume': pl.Float64,
        'taker_buy_quote_volume': pl.Float64,
        'ignore': pl.String,
    }

    columns: list[str | pl.Expr] = [
        'timestamp',
        pl.lit(pair).alias('pair'),
        'open',
        'high',
        'low',
        'close',
        'n_trades',
        'base_volume',
        'quote_volume',
        'taker_buy_base_volume',
        'taker_buy_quote_volume',
    ]

    return _process(url=url, raw_schema=raw_schema, columns=columns)


def _process(
    url: str,
    raw_schema: dict[str, pl.DataType | type[pl.DataType]],
    columns: list[str | pl.Expr],
) -> pl.DataFrame:
    import polars as pl

    datetime_column = (
        pl.when(pl.col.timestamp >= 1230796800 * 1_000_000)
        .then(pl.col.timestamp.cast(pl.Datetime('us')))
        .when(pl.col.timestamp >= 1230796800 * 1_000)
        .then((1_000 * pl.col.timestamp).cast(pl.Datetime('us')))
        .when(pl.col.timestamp >= 1230796800)
        .then((1_000_000 * pl.col.timestamp).cast(pl.Datetime('us')))
    )

    return (
        absorb.ops.download_csv_zip_to_dataframe(
            url, polars_kwargs={'schema': raw_schema, 'has_header': False}
        )
        .with_columns(datetime_column)
        .select(columns)
    )



from __future__ import annotations

import typing

import absorb
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace


def upload_command(args: Namespace) -> dict[str, typing.Any]:
    # determine tables to upload
    tables = cli_parsing._parse_datasets(args)

    # determine bucket to upload to
    bucket = cli_parsing._parse_bucket(args)

    # print summary
    print(
        'uploading '
        + str(len(tables))
        + ' tables to bucket '
        + str(bucket['bucket_name'])
    )
    for table in tables:
        print('- ' + table.full_name())

    # exit early if dry
    if args.dry:
        return {}

    # upload tables
    absorb.ops.upload_tables_to_bucket(tables=tables, bucket=bucket)

    return {}



from __future__ import annotations

from . import common


class BeaconComitteeEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_beacon_committee'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconAttestationEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_events_attestation'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconBlobSidecarEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_events_blob_sidecar'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconBlockEventsV1(common.XatuTable):
    datatype = 'beacon_api_eth_v1_events_block'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconChainReorgEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_events_chain_reorg'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconContributionAndProofEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_events_contribution_and_proof'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconFinalizedCheckpointEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_events_finalized_checkpoint'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconHeadEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_events_head'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconVoluntaryExitEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_events_voluntary_exit'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconValidatorAttestationEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_validator_attestation_data'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconBlockEventsV2(common.XatuTable):
    datatype = 'beacon_api_eth_v2_beacon_block'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconProposerDutyEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v1_proposer_duty'
    source = 'beacon_api'
    index_type = 'hour'


class BeaconValidatorEvents(common.XatuTable):
    datatype = 'beacon_api_eth_v3_validator_block'
    source = 'beacon_api'
    index_type = 'hour'



from __future__ import annotations

import typing
import absorb

if typing.TYPE_CHECKING:
    import polars as pl

url_template = 'https://mempool-dumpster.flashbots.net/ethereum/mainnet/{year}-{month:02}/{year}-{month:02}-{day:02}.parquet'


class Transactions(absorb.Table):
    source = 'mempool_dumpster'
    url = 'https://github.com/flashbots/mempool-dumpster'
    description = 'Archive of the Ethereum mempool collected by Flashbots'
    write_range = 'append_only'
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'inclusion_block_number': pl.Int64,
            'inclusion_delay': pl.Int64,
            'inclusion_timestamp': pl.Datetime('ms'),
            'timestamp': pl.Datetime('ms'),
            'hash': pl.Binary,
            'from': pl.Binary,
            'to': pl.Binary,
            'value': pl.Float64,
            'nonce': pl.Int64,
            'gas': pl.Int64,
            'gas_price': pl.Float64,
            'gas_tip_cap': pl.Float64,
            'gas_fee_cap': pl.Float64,
            'data_size': pl.Int64,
            'data_byte': pl.Binary,
            'sources': pl.List(pl.String),
            'raw_transaction': pl.String,
            'chain_id': pl.String,
        }

    source = 'mempool_dumpster'
    renamed = {
        'hash': 'transaction_hash',
        'gas': 'gas_used',
        'from': 'from_address',
        'to': 'to_address',
        'chainId': 'chain_id',
        # 'txType': 'tx_type',
        'gasPrice': 'gas_price',
        'gasTipCap': 'max_priority_fee_per_gas',
        'gasFeeCap': 'max_fee_per_gas',
        'dataSize': 'n_data_bytes',
        'data4Bytes': 'data_4byte',
        'rawTx': 'raw_transaction',
        'timestamp': 'submission_timestamp',
        'includedAtBlockHeight': 'inclusion_block_number',
        'includedBlockTimestamp': 'inclusion_timestamp',
        'inclusionDelayMs': 'inclusion_delay',
    }

    binary_columns = [
        'hash',
        'from',
        'to',
        'data4Bytes',
        'rawTx',
    ]

    def get_available_range(self) -> absorb.Coverage:
        import datetime

        current = datetime.datetime.now()
        current = datetime.datetime(
            year=current.year, month=current.month, day=current.day
        )
        current = current + datetime.timedelta(days=1)
        initial = datetime.datetime(year=2023, month=8, day=8)
        while current > initial:
            url = url_template.format(
                year=current.year, month=current.month, day=current.day
            )
            if absorb.ops.does_remote_file_exist(url):
                break
            current -= datetime.timedelta(days=1)
        return (initial, current)

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        url = url_template.format(
            year=chunk.year,  # type: ignore
            month=chunk.month,  # type: ignore
            day=chunk.day,  # type: ignore
        )
        return absorb.ops.download_parquet_to_dataframe(url)

    # @classmethod
    # def scan(
    #     dataset: absorb.TableReference,
    #     *,
    #     start_time: tooltime.Timestamp | None = None,
    #     end_time: tooltime.Timestamp | None = None,
    #     root_dir: str | None = None,
    #     flat: bool | None = None,
    #     extra_kwargs: dict[str, typing.Any] | None = None,
    # ) -> pl.LazyFrame:
    #     lf = super().scan()

    #     rename: bool = extra_kwargs.get('rename', True)
    #     reorder: bool = extra_kwargs.get('reorder', True)
    #     deduplicate: bool = extra_kwargs.get('deduplicate', True)
    #     columns: typing.Sequence[str] | None = extra_kwargs.get('columns', None)

    #     lf = lf.with_columns(
    #         value=pl.col.value.cast(pl.Float64),
    #         nonce=pl.col.nonce.cast(pl.Int64),
    #         gas=pl.col.gas.cast(pl.Float64),
    #         gasPrice=pl.col.gasPrice.cast(pl.Float64),
    #         gasTipCap=pl.col.gasTipCap.cast(pl.Float64),
    #         gasFeeCap=pl.col.gasFeeCap.cast(pl.Float64),
    #     )

    #     # select and reorder columns
    #     if reorder:
    #         lf = lf.select(reordered)
    #     if columns is not None:
    #         renamed_reverse = {v: k for k, v in renamed.items()}
    #         columns = [
    #             renamed_reverse.get(column, column) for column in columns
    #         ]
    #         lf = lf.select(columns)

    #     # deduplicate redundant transactions
    #     if deduplicate:
    #         if columns is None:
    #             columns = lf.collect_schema().names()
    #         column_kwargs = {}
    #         for column in columns:
    #             if column != 'hash':
    #                 column_kwargs[column] = pl.col(column).min()
    #         lf = lf.group_by('hash').agg(**column_kwargs)

    #     # rename columns
    #     if rename:
    #         if columns is not None:
    #             use_rename = {k: v for k, v in renamed.items() if k in columns}
    #         else:
    #             use_rename = renamed
    #         lf = lf.rename(use_rename)

    #     return lf


def load_block_stats(
    *, start_block: int | None = None, end_block: int | None = None
) -> pl.DataFrame:
    import cryo_manager  # type: ignore
    import polars as pl

    return (  # type: ignore
        cryo_manager.scan(
            'blocks',
            network='ethereum',
            include_timestamps=True,
        )
        .sort('block_number')
        .select(
            'block_number',
            'timestamp',
            'base_fee_per_gas',
            'gas_used',
            'gas_limit',
            utilization=pl.col.gas_used / pl.col.gas_limit,
        )
        .with_columns(prev_utilization=pl.col.utilization.shift(1))
        .collect()
    )


def join_block_stats(txs: pl.DataFrame) -> pl.DataFrame:
    import polars as pl

    blocks = load_block_stats(
        start_block=txs['inclusion_block_number'].min(),  # type: ignore
        end_block=txs['inclusion_block_number'].max(),  # type: ignore
    )
    blocks = blocks[
        'block_number', 'base_fee_per_gas', 'utilization', 'prev_utilization'
    ]

    return (
        txs.join(
            blocks,
            left_on='inclusion_block_number',
            right_on='block_number',
            how='left',
            coalesce=True,
        )
        .sort('inclusion_block_number')
        .with_columns(priority_fee=pl.col.gas_price - pl.col.base_fee_per_gas)
    )



from __future__ import annotations

from . import common


class ExecutionBlocks(common.XatuTable):
    datatype = 'canonical_execution_block'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionTransactions(common.XatuTable):
    datatype = 'canonical_execution_transaction'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionTraces(common.XatuTable):
    datatype = 'canonical_execution_traces'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionLogs(common.XatuTable):
    datatype = 'canonical_execution_logs'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionContracts(common.XatuTable):
    datatype = 'canonical_execution_contracts'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionFourByteCounts(common.XatuTable):
    datatype = 'canonical_execution_four_byte_counts'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionAddressAppearances(common.XatuTable):
    datatype = 'canonical_execution_address_appearances'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionBalanceDiffs(common.XatuTable):
    datatype = 'canonical_execution_balance_diffs'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionBalanceReads(common.XatuTable):
    datatype = 'canonical_execution_balance_reads'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionErc20Transfers(common.XatuTable):
    datatype = 'canonical_execution_erc20_transfers'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionErc721Transfers(common.XatuTable):
    datatype = 'canonical_execution_erc721_transfers'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionNativeTransfers(common.XatuTable):
    datatype = 'canonical_execution_native_transfers'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionNonceDiffs(common.XatuTable):
    datatype = 'canonical_execution_nonce_diffs'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionNonceReads(common.XatuTable):
    datatype = 'canonical_execution_nonce_reads'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionStorageDiffs(common.XatuTable):
    datatype = 'canonical_execution_storage_diffs'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}


class ExecutionStorageReads(common.XatuTable):
    datatype = 'canonical_execution_storage_reads'
    source = 'canonical_execution'
    index_type = {'type': 'number_range', 'number_interval': 1000}



from __future__ import annotations

import typing

import absorb
from . import io

if typing.TYPE_CHECKING:
    import polars as pl


def query(
    table: absorb.TableReference,
    *,
    update: bool = False,
    collect_if_missing: bool = True,
    scan_kwargs: dict[str, typing.Any] | None = None,
) -> pl.LazyFrame:
    table = absorb.Table.instantiate(table)

    # check if collected
    if not table.is_collected():
        if collect_if_missing or update:
            table.collect()
        else:
            raise Exception(
                f'Table {table.source}.{table.name()} is not collected.'
            )
    elif update:
        table.collect()

    # scan the table
    return io.scan(table, scan_kwargs=scan_kwargs)


def sql_query(
    sql: str,
    *,
    backend: typing.Literal['absorb', 'dune', 'snowflake'] = 'absorb',
) -> pl.LazyFrame:
    if backend == 'absorb':
        # create table context
        context = create_sql_context()

        # modify query to allow dots in names
        for table in context.tables():
            if '.' in table and table in sql:
                sql = sql.replace(table, '"' + table + '"')

        return context.execute(sql)  # type: ignore
    elif backend == 'dune':
        import spice

        return spice.query(sql).lazy()
    elif backend == 'snowflake':
        import garlic

        return garlic.query(sql).lazy()
    else:
        raise Exception('invalid backend: ' + backend)


def create_sql_context(
    *,
    tracked_tables: bool = True,
    collected_tables: bool = True,
) -> pl.SQLContext[typing.Any]:
    import polars as pl

    # decide which tables to include
    all_tables = []
    if tracked_tables:
        all_tables += absorb.ops.get_tracked_tables()
    if collected_tables:
        all_tables += absorb.ops.get_collected_tables()

    # index tables by full name
    tables_by_name = {}
    for table_dict in all_tables:
        name = table_dict['source_name'] + '.' + table_dict['table_name']
        if name not in tables_by_name:
            tables_by_name[name] = absorb.scan(table_dict)

    # create context
    return pl.SQLContext(**tables_by_name)  # type: ignore



"""python interface for interacting with flashbots mempool dumpster"""

from .errors import *
from .table import Table
from . import ops
from .ops import (
    scan,
    load,
    query,
    sql_query,
    get_available_range,
    get_collected_range,
)

import typing

if typing.TYPE_CHECKING:
    from .annotations import *


__version__ = '0.3.0'



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl


def does_remote_file_exist(url: str) -> bool:
    import requests

    try:
        response = requests.head(url, allow_redirects=True)
        # Check if status code is 200 (OK) and content-length exists
        if response.status_code == 200 and 'content-length' in response.headers:
            return True
        return False
    except requests.RequestException:
        return False


def download_file(*, url: str, path: str) -> None:
    raise NotImplementedError()


def download_parquet_to_dataframe(url: str) -> pl.DataFrame:
    import io
    import requests
    import polars as pl

    try:
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception(
                f'Failed to download: HTTP status code {response.status_code}'
            )
        parquet_buffer = io.BytesIO(response.content)
        return pl.read_parquet(parquet_buffer)
    except Exception as e:
        raise Exception(f'Error processing parquet file: {str(e)}')


def download_csv_gz_to_dataframe(
    url: str, *, polars_kwargs: dict[str, typing.Any] | None = None
) -> pl.DataFrame:
    import io
    import gzip
    import requests
    import polars as pl

    try:
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception(
                f'Failed to download: HTTP status code {response.status_code}'
            )
        csv_buffer = io.StringIO(
            gzip.decompress(response.content).decode('utf-8')
        )
        if polars_kwargs is None:
            polars_kwargs = {}
        return pl.read_csv(csv_buffer, **polars_kwargs)
    except Exception as e:
        raise Exception(f'Error processing csv.gz file: {str(e)}')


def download_csv_zip_to_dataframe(
    url: str, *, polars_kwargs: dict[str, typing.Any] | None = None
) -> pl.DataFrame:
    import io
    import zipfile
    import requests
    import polars as pl

    try:
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception(
                f'Failed to download: HTTP status code {response.status_code}'
            )
        zip_buffer = io.BytesIO(response.content)
        with zipfile.ZipFile(zip_buffer, 'r') as z:
            csv_filename = [f for f in z.namelist() if f.endswith('.csv')][0]
            with z.open(csv_filename) as csv_file:
                csv_buffer = io.StringIO(csv_file.read().decode('utf-8'))
                if polars_kwargs is None:
                    polars_kwargs = {}
                return pl.read_csv(csv_buffer, **polars_kwargs)
    except Exception as e:
        raise Exception(f'Error processing csv.zip file: {str(e)}')



from __future__ import annotations

import typing

import absorb
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def path_command(args: Namespace) -> dict[str, Any]:
    path = _get_path(args)
    print(path)
    return {}


def _get_path(args: Namespace) -> str:
    if args.dataset is None:
        return absorb.ops.get_absorb_root(warn=False)
    elif args.glob:
        tracked_dataset = cli_parsing._parse_datasets(args)[0]
        instance = absorb.Table.instantiate(tracked_dataset)
        return instance.get_glob(warn=False)
    elif '.' in args.dataset:
        source, table = args.dataset.split('.')
        return absorb.ops.get_table_dir(source=source, table=table, warn=False)
    else:
        return absorb.ops.get_source_dir(args.dataset, warn=False)



# https://docs.growthepie.xyz/api

from __future__ import annotations

import absorb

import typing

if typing.TYPE_CHECKING:
    import polars as pl


class Metrics(absorb.Table):
    description = 'On-chain metrics for Ethereum and its rollups'
    url = 'https://www.growthepie.com/'
    source = 'growthepie'
    write_range = 'overwrite_all'
    index_type = 'day'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'date': pl.Datetime(time_unit='us', time_zone=None),
            'network': pl.String,
            'market_cap_usd': pl.Float64,
            'market_cap_eth': pl.Float64,
            'txcount': pl.Float64,
            'aa_last7d': pl.Float64,
            'txcosts_median_eth': pl.Float64,
            'daa': pl.Float64,
            'gas_per_second': pl.Float64,
            'fees_paid_eth': pl.Float64,
            'profit_usd': pl.Float64,
            'app_fees_eth': pl.Float64,
            'fees_paid_usd': pl.Float64,
            'txcosts_median_usd': pl.Float64,
            'stables_mcap_eth': pl.Float64,
            'fdv_eth': pl.Float64,
            'stables_mcap': pl.Float64,
            'costs_l1_usd': pl.Float64,
            'costs_total_usd': pl.Float64,
            'rent_paid_usd': pl.Float64,
            'app_fees_usd': pl.Float64,
            'tvl': pl.Float64,
            'tvl_eth': pl.Float64,
            'fdv_usd': pl.Float64,
            'costs_blobs_usd': pl.Float64,
            'rent_paid_eth': pl.Float64,
            'costs_blobs_eth': pl.Float64,
            'profit_eth': pl.Float64,
            'costs_l1_eth': pl.Float64,
            'costs_total_eth': pl.Float64,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import requests
        import polars as pl

        url = 'https://api.growthepie.xyz/v1/fundamentals_full.json'
        response = requests.get(url)
        data = response.json()
        return (
            pl.DataFrame(data)
            .with_columns(date=pl.col.date.str.to_date().cast(pl.Datetime))
            .rename({'origin_key': 'network'})
            .pivot(on='metric_key', index=['date', 'network'], values='value')
            .sort('date', 'network')
        )

    def get_available_range(self) -> absorb.Coverage:
        import datetime
        import requests
        import polars as pl

        first = datetime.datetime(year=2021, month=6, day=1)

        url = 'https://api.growthepie.xyz/v1/fundamentals.json'
        response = requests.get(url)
        data = response.json()
        last_str: str = pl.DataFrame(data)['date'].max()  # type: ignore
        last = datetime.datetime.strptime(last_str, '%Y-%m-%d')

        return (first, last)



from __future__ import annotations

import typing
import polars as pl
import absorb


class BaseQuery(absorb.Table):
    source = 'dune'
    url = 'https://dune.com/'
    name_template = [
        '{name}',
        'query_{query_id}',
    ]
    required_packages = ['dune_spice >= 0.2.6']
    required_credentials = ['DUNE_API_KEY']

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import spice

        query = self.parameters['query']
        spice_kwargs = self.parameters['spice_kwargs']
        spice_kwargs['limit'] = 0
        return dict(spice.query(query, **spice_kwargs).schema)


class FullQuery(BaseQuery):
    """collect the full output of a query"""

    description = 'Dune query'
    write_range = 'overwrite_all'
    parameter_types = {
        'name': str,
        'query_id': str,
        'spice_kwargs': dict[str, typing.Any],
    }
    required_packages = ['dune_spice >= 0.2.6']
    required_credentials = ['DUNE_API_KEY']

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import spice

        query = self.parameters['query']
        spice_kwargs = self.parameters['spice_kwargs']
        return spice.query(
            query, poll=True, include_execution=False, **spice_kwargs
        )


class AppendOnlyQuery(BaseQuery):
    """collect the output of a query, time-partitioned"""

    description = 'Dune query has an append-only structure'
    write_range = 'append_only'
    parameter_types = {
        'name': str,
        'query_id': str,
        'spice_kwargs': dict[str, typing.Any],
        'range_parameters': list[str],
    }
    required_packages = ['dune_spice >= 0.2.6']
    required_credentials = ['DUNE_API_KEY']

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import spice

        query = self.parameters['query']
        spice_kwargs = self.parameters['spice_kwargs']
        spice_kwargs.setdefault('parameters', {})
        self.parameters.update(chunk)  # type: ignore
        return spice.query(
            query, poll=True, include_execution=False, **spice_kwargs
        )


class CexLabels(absorb.Table):
    source = 'dune'
    description = 'CEX labels for EVM and Solana addresses'
    url = 'https://dune.com/'
    write_range = 'overwrite_all'
    required_packages = ['dune_spice >= 0.2.6']
    required_credentials = ['DUNE_API_KEY']
    index_type = 'id'

    def get_schema(self) -> dict[str, pl.DataType | type[pl.DataType]]:
        import polars as pl

        return {
            'address': pl.String,
            'cex_name': pl.String,
            'distinct_name': pl.String,
            'added_by': pl.String,
            'added_date': pl.String,
            'ecosystem': pl.String,
        }

    def collect_chunk(self, chunk: absorb.Chunk) -> absorb.ChunkData | None:
        import spice

        evm_cex_query = 'https://dune.com/queries/3237025'
        solana_cex_query = 'https://dune.com/queries/5124188'
        evm_cexes = spice.query(evm_cex_query).with_columns(
            ecosystem=pl.lit('EVM')
        )
        solana_cexes = (
            spice.query(solana_cex_query)
            .drop('blockchain')
            .with_columns(ecosystem=pl.lit('solana'))
        )
        return pl.concat([evm_cexes, solana_cexes])


def get_tables() -> list[type[absorb.Table]]:
    return [CexLabels]



from __future__ import annotations

import typing

import absorb


bullet_styles = {
    'key_style': 'white bold',
    'bullet_style': 'green',
    'colon_style': 'green',
}


def print_bullet(
    key: str | None, value: str | None, **kwargs: typing.Any
) -> None:
    import toolstr

    toolstr.print_bullet(key=key, value=value, **kwargs, **bullet_styles)


def format_coverage(
    coverage: absorb.Coverage | None, index_type: absorb.IndexType | None
) -> str:
    if coverage is None:
        return 'None'
    if isinstance(coverage, tuple):
        start, end = coverage
        return (
            format_chunk(start, index_type)
            + '_to_'
            + format_chunk(end, index_type)
        )
    elif isinstance(coverage, list):
        start = min(coverage)
        end = max(coverage)
        return (
            format_chunk(start, index_type)
            + '_to_'
            + format_chunk(end, index_type)
        )
    elif isinstance(coverage, dict):
        raise NotImplementedError()
    else:
        raise Exception()


def format_chunk(
    chunk: absorb.Chunk, index_type: absorb.IndexType | None
) -> str:
    if chunk is None:
        return '-'
    if index_type is None:
        return '-'

    if index_type == 'hour':
        return chunk.strftime('%Y-%m-%d--%H-%M-%S')  # type: ignore
    elif index_type == 'day':
        return chunk.strftime('%Y-%m-%d')  # type: ignore
    elif index_type == 'week':
        return chunk.strftime('%Y-%m-%d')  # type: ignore
    elif index_type == 'month':
        return chunk.strftime('%Y-%m')  # type: ignore
    elif index_type == 'quarter':
        if chunk.month == 1 and chunk.day == 1:  # type: ignore
            quarter = 1
        elif chunk.month == 4 and chunk.day == 1:  # type: ignore
            quarter = 2
        elif chunk.month == 7 and chunk.day == 1:  # type: ignore
            quarter = 4
        elif chunk.month == 10 and chunk.day == 1:  # type: ignore
            quarter = 4
        else:
            raise Exception('invalid quarter timestamp')
        return chunk.strftime('%Y-Q') + str(quarter)  # type: ignore
    elif index_type == 'year':
        return chunk.strftime('%Y')  # type: ignore
    elif index_type == 'timestamp':
        return chunk.strftime('%Y-%m-%d--%H-%M-%S')  # type: ignore
    elif index_type == 'timestamp_range':
        import datetime

        t_start: datetime.datetime
        t_end: datetime.datetime
        t_start, t_end = chunk  # type: ignore
        return (
            t_start.strftime('%Y-%m-%d--%H-%M-%S')
            + '_to_'
            + t_end.strftime('%Y-%m-%d--%H-%M-%S')
        )
    elif index_type == 'number':
        width = 10
        template = '%0' + str(width) + 'd'
        return template % chunk
    elif index_type == 'number_range':
        width = 10
        template = '%0' + str(width) + 'd'
        start, end = chunk  # type: ignore
        return (template % start) + '_to_' + (template % end)
    elif index_type == 'id':
        return chunk  # type: ignore
    elif index_type == 'id_list':
        return '_'.join(chunk)  # type: ignore
    elif index_type is None:
        return str(chunk)
    else:
        raise Exception('invalid chunk range format: ' + str(index_type))



from __future__ import annotations

import typing


def validate_table_dict(metadata: typing.Any) -> list[str]:
    if not isinstance(metadata, dict):
        return ['metadata is not a dictionary']

    errors = []
    for key, value_type in {
        'source_name': str,
        'table_name': str,
        'table_class': str,
        'table_version': str,
        'parameters': dict,
    }.items():
        if key not in metadata:
            errors.append(f'missing required key: {key}')
            continue
        if not isinstance(metadata[key], value_type):
            errors.append(
                f'key {key} is not of type {value_type.__name__}, got {type(metadata[key]).__name__}'
            )

    return errors
